Package: unifiedml
Type: Package
Title: Unified Interface for Machine Learning Models
Version: 0.3.0
Date: 2026-05-05
Authors@R: 
    person("T.", "Moudiki", 
           email = "thierry.moudiki@gmail.com",
           role = c("aut", "cre"))
Maintainer: T. Moudiki <thierry.moudiki@gmail.com>
Description: Provides a unified R6-based interface for various machine learning models with automatic interface detection, consistent cross-validation, model interpretations via numerical derivatives, and visualization. Supports both regression and classification tasks with any model function that follows R's standard modeling conventions (formula or matrix interface).
License: MIT + file LICENSE
URL: https://github.com/Techtonique/unifiedml
BugReports: https://github.com/Techtonique/unifiedml/issues
Depends: R (>= 3.5.0), doParallel, R6, foreach
Imports: Rcpp (>= 1.1.0)
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, glmnet, randomForest,
        e1071, covr, spelling, MASS, nnet, caret
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.2
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-05-05 06:54:30 UTC; t
Author: T. Moudiki [aut, cre]
Repository: CRAN
Date/Publication: 2026-05-05 07:10:02 UTC
Built: R 4.5.2; ; 2026-05-05 11:31:11 UTC; unix

import(R6)
import(Rcpp)
import(foreach)
import(doParallel)
importFrom(stats, predict)
importFrom(stats, family)
importFrom("stats", "model.matrix")
importFrom(utils, txtProgressBar)
export(Model)
export(cross_val_score)
export(formula_to_matrix)
export(matrix_to_formula)
export(train_test_split)
export(benchmark)
export(extract_probabilities)

# unifiedml

> A Unified Machine Learning Interface for R

[![](https://cranlogs.r-pkg.org/badges/unifiedml)](https://cran.r-project.org/package=unifiedml)
[![](https://cranlogs.r-pkg.org/badges/grand-total/unifiedml)](https://cranlogs.r-pkg.org/badges/grand-total/unifiedml)
[![Documentation](https://img.shields.io/badge/documentation-is_here-green)](http://docs.techtonique.net/unifiedml/index.html)

## Overview

`unifiedml` provides a (work in progress) consistent, sklearn-like interface for various (any) machine learning models in R. 

It eliminates the need to remember different function signatures across packages by **automatically detecting the appropriate interface** (formula vs matrix) and **task type** (regression vs classification).

### Key Features

For now: 

- **Automatic Task Detection**: Automatically detects regression vs classification based on response variable type (numeric → regression, factor → classification)
- **Universal Interface**: Works seamlessly with `glmnet`, `randomForest`, `e1071::svm`, and other popular ML packages with formula or matrix interface
- **Built-in Cross-Validation**: Consistent `cross_val_score()` function with automatic metric selection
- **Model Interpretability**: Numerical derivatives and statistical significance testing via `summary()`
- **Partial Dependence Plots**: Visualize feature effects with `plot()`
- **Method Chaining**: Clean, pipeable syntax with R6 classes

## Installation

- From CRAN: 

  ```R
  install.packages("unifiedml")
  ```
  
- From Github (development version):
  
  ```R
  # Install from GitHub (development version, for now)
  devtools::install_github("Techtonique/unifiedml")
  ```

## Quick Start

### Regression Example

```R
library(unifiedml)
library(glmnet)

# Prepare data
data(mtcars)
X <- as.matrix(mtcars[, -1])
y <- mtcars$mpg  # numeric → automatic regression

# Fit model
mod <- Model$new(glmnet::glmnet)
mod$fit(X, y, alpha = 0, lambda = 0.1)

# Make predictions
predictions <- mod$predict(X)

# Get model summary with feature importance
mod$summary()

# Visualize partial dependence
mod$plot(feature = 1)

# Cross-validation (automatically uses RMSE for regression)
cv_scores <- cross_val_score(mod, X, y, cv = 5)
cat("Mean RMSE:", mean(cv_scores), "\n")
```

### Classification Example

```R
library(randomForest)

# Prepare data
data(iris)
X <- as.matrix(iris[, 1:4])
y <- iris$Species  # factor → automatic classification

# Fit model
mod <- Model$new(randomForest::randomForest)
mod$fit(X, y, ntree = 100)

# Make predictions
predictions <- mod$predict(X)

# Get model summary
mod$summary()

# Cross-validation (automatically uses accuracy for classification)
cv_scores <- cross_val_score(mod, X, y, cv = 5)
cat("Mean Accuracy:", mean(cv_scores), "\n")
```

## Core Functionality

### The Model Class

The `Model` R6 class provides a unified interface for any machine learning function:

```R
# Create a model wrapper
mod <- Model$new(model_function)

# Fit the model (task type auto-detected from y)
mod$fit(X, y, ...)

# Make predictions
predictions <- mod$predict(X_new)

# Get interpretable summary
mod$summary(h = 0.01, alpha = 0.05)

# Visualize feature effects
mod$plot(feature = 1)

# Print model info
mod$print()
```

### Cross-Validation

The `cross_val_score()` function provides consistent k-fold cross-validation:

```R
# Automatic metric selection based on task
scores <- cross_val_score(mod, X, y, cv = 5)

# Specify custom metric
scores <- cross_val_score(mod, X, y, cv = 10, scoring = "mae")

# Available metrics:
# - Regression: "rmse" (default), "mae"
# - Classification: "accuracy" (default), "f1"
```

### Model Interpretability

The `summary()` method uses numerical derivatives to assess feature importance:

```R
mod$summary()
# Output:
# Model Summary - Numerical Derivatives
# ======================================
# Task: regression
# Samples: 150 | Features: 4
# 
# Feature         Mean_Derivative  Std_Error  t_value  p_value  Significance
# Sepal.Length    0.523           0.042       12.45    < 0.001  ***
# Sepal.Width    -0.234           0.038       -6.16    < 0.001  ***
# ...
```

## Supported Models

`unifiedml` automatically detects the appropriate interface for:

- **glmnet**: Ridge, Lasso, Elastic Net regression and classification
- **randomForest**: Random forest for regression and classification
- **e1071::svm**: Support Vector Machines
- **Any model** with either formula (`y ~ .`) or matrix (`x, y`) interface

## Automatic Task Detection

The package automatically determines the task type:

```R
# Regression (numeric y)
y_reg <- c(1.2, 3.4, 5.6, ...)
mod$fit(X, y_reg)  # → task = "regression"

# Classification (factor y)
y_class <- factor(c("A", "B", "A", ...))
mod$fit(X, y_class)  # → task = "classification"
```

## Advanced Features

### Partial Dependence Plots

```R
# Visualize how feature j affects predictions
mod$plot(feature = 3, n_points = 100)
```

### Model Cloning

```R
# Create independent copy for parallel processing
mod_copy <- mod$clone_model()
```

## Examples

See the package vignette for comprehensive examples:

```R
vignette("introduction", package = "unifiedml")
```

## Why unifiedml?

Traditional R modeling requires remembering different interfaces:

```R
# Different interfaces = cognitive overhead
glmnet(x = X, y = y, ...)               # matrix interface
randomForest(y ~ ., data = df, ...)     # formula interface
svm(x = X, y = y, ...)                  # matrix interface
```

With `unifiedml`, it's always the same:

```R
# One interface to rule them all
Model$new(glmnet)$fit(X, y, ...)
Model$new(randomForest)$fit(X, y, ...)
Model$new(svm)$fit(X, y, ...)
```

## Contributing

Contributions are welcome, feel free to submit a Pull Request.

## License

The Clear BSD License - see LICENSE file for details.

## Citation

If you use this package in your research, please cite:

```
@Manual{unifiedml,
  title = {unifiedml: Unified Machine Learning Interface for R},
  author = {T. Moudiki},
  year = {2025},
  note = {R package version 0.1.0}
}
```

## ----fig.width=7--------------------------------------------------------------
library(unifiedml) # this package
library(randomForest)
library(e1071)
library(glmnet)

# ------------------------------------------------------------
# REGRESSION EXAMPLES
# ------------------------------------------------------------

cat("\n=== REGRESSION EXAMPLES ===\n\n")

# Example 1: Synthetic data (numeric y → automatic regression)
 
X <- matrix(rnorm(100), ncol = 4)
y <- 2*X[,1] - 1.5*X[,2] + rnorm(25)  # numeric -> regression
 
mod <- Model$new(glmnet::glmnet)
mod$fit(X, y, alpha = 0, lambda = 0.1)
mod$print()
mod$summary(h = 0.01)
print(head(mod$predict(X)))
(cv_scores <- cross_val_score(mod, X, y, cv = 5))  # auto-uses RMSE
mean(cv_scores)  # Average RMSE


# Example 2: Random Forest Regression
cat("\n2. Random Forest Regression - Auto-detected: Regression\n")
# randomForest regression
set.seed(123)
X <- MASS::Boston[, -ncol(MASS::Boston)]
y <- MASS::Boston$medv
mod2 <- Model$new(randomForest::randomForest)  # No task parameter!
mod2$fit(X, y, ntree = 50)
mod2$print()
cat("\n")
mod2$summary(h = 0.01)
(cv2 <- cross_val_score(mod2, X, y, cv = 5L)) 
print(head(mod2$predict(X)))

# ------------------------------------------------------------
# CLASSIFICATION EXAMPLES
# ------------------------------------------------------------

cat("\n\n=== CLASSIFICATION EXAMPLES ===\n\n")

# Example: Iris dataset (factor y → automatic classification)
data(iris)

# Binary classification with factor
cat("3. Binary Classification with Factor Response\n")
iris_binary <- iris[iris$Species %in% c("setosa", "versicolor"), ]
X_binary <- iris_binary[, 1:4]
y_binary <- as.factor(as.character(iris_binary$Species))  # factor → classification


mod4 <- Model$new(randomForest::randomForest)  # No task parameter!
mod4$fit(X_binary, y_binary, ntree = 50)
mod4$print()
print(head(mod4$predict(X_binary)))
print(head(mod4$predict(X_binary, type="prob")))
(cv4 <- cross_val_score(mod4, X_binary, y_binary, cv = 5L))  # Auto-uses accuracy
cat("\nMean Accuracy:", mean(cv4), "\n")


mod4 <- Model$new(nnet::nnet) 
mod4$fit(X_binary, y_binary, size=50, trace=FALSE)
print(head(mod4$predict(X_binary, type="class")))
(cv4 <- cross_val_score(mod4, X_binary, y_binary, fit_params=list(size=50L, 
                        cv = 5L, type="class", trace=FALSE)))
cat("\nMean Accuracy:", mean(cv4), "\n")


mod4 <- Model$new(e1071::svm)  # No task parameter!
mod4$fit(X_binary, y_binary)
mod4$print()
print(head(mod4$predict(X_binary)))
(cv4 <- cross_val_score(mod4, X_binary, y_binary, cv = 5L))  # Auto-uses accuracy
cat("\nMean Accuracy:", mean(cv4), "\n")


# Multi-class classification
cat("4. Multi-class Classification\n")
X_multi <- iris[, 1:4]
y_multi <- iris$Species  # factor with 3 levels → multi-class classification

mod4 <- Model$new(randomForest::randomForest)  # No task parameter!
mod4$fit(X_multi, y_multi, ntree = 50)
mod4$print()
print(head(mod4$predict(X_multi)))
print(head(mod4$predict(X_binary, type="prob")))
(cv4 <- cross_val_score(mod4, X_multi, y_multi, cv = 5L))  # Auto-uses accuracy
cat("\nMean Accuracy:", mean(cv4), "\n")


mod4 <- Model$new(nnet::nnet) 
mod4$fit(X_multi, y_multi, size=50, trace=FALSE)
print(head(mod4$predict(X_multi, type="class")))
(cv4 <- cross_val_score(mod4, X_multi, y_multi, fit_params=list(size=50L, 
                        cv = 5L, type="class", trace=FALSE)))
cat("\nMean Accuracy:", mean(cv4), "\n")


mod4 <- Model$new(e1071::svm)  # No task parameter!
mod4$fit(X_multi, y_multi, kernel="radial")
mod4$print()
print(head(mod4$predict(X_multi)))
(cv4 <- cross_val_score(mod4, X_multi, y_multi, cv = 5L))  # Auto-uses accuracy
cat("\nMean Accuracy:", mean(cv4), "\n")


mod4 <- Model$new(e1071::svm)  # No task parameter!
mod4$fit(X_multi, y_multi, kernel="polynomial", probability=TRUE)
mod4$print()
print(head(mod4$predict(X_multi)))
(cv4 <- cross_val_score(mod4, X_multi, y_multi, cv = 5L))  # Auto-uses accuracy
cat("\nMean Accuracy:", mean(cv4), "\n")

## ----fig.width=7--------------------------------------------------------------
library(unifiedml) # this package
require(caret)

set.seed(123)

## ----fig.width=7--------------------------------------------------------------
iris_binary <- iris[iris$Species %in% c("setosa", "versicolor"), ]
X_binary <- iris_binary[, 1:4]
y_binary <- as.factor(as.character(iris_binary$Species))  # factor → classification

datasplit <- unifiedml::train_test_split(X_binary, y_binary, 
                                         test_size = 0.3, seed = 42)

mod <- Model$new(caret::train) 
mod$fit(datasplit$X_train, datasplit$y_train,
        method = "rf",
        trControl = caret::trainControl(method = "none"))
print(head(mod$predict(datasplit$X_test)))
print(head(mod$predict(datasplit$X_test, type="prob")))

## ----fig.width=7--------------------------------------------------------------
X <- iris[, 1:4]
y <- iris$Species  # factor → classification

datasplit <- unifiedml::train_test_split(X, y, 
                                         test_size = 0.3, seed = 42)

mod <- Model$new(caret::train) 
mod$fit(datasplit$X_train, datasplit$y_train,
        method = "glmnet",
        tuneGrid = data.frame(alpha = 0,  # ridge regression
                               lambda = 0.01),  # fixed lambda
        trControl = caret::trainControl(method = "none"))
print(head(mod$predict(datasplit$X_test)))
print(head(mod$predict(datasplit$X_test, type="prob")))

(cv <- cross_val_score(mod, datasplit$X_train, datasplit$y_train, cv = 5L, 
                       fit_params=list(method = "glmnet",
                       tuneGrid = data.frame(alpha = 0,  # ridge regression
                       lambda = 0.01),  # fixed lambda
                       trControl = caret::trainControl(method = "none"))))  # Auto-uses accuracy
cat("\nMean Accuracy:", mean(cv), "\n")


(cv <- cross_val_score(mod, datasplit$X_train, datasplit$y_train, cv = 5L, 
                       fit_params=list(method = "glmnet",
                       tuneGrid = data.frame(alpha = 0.5,  # ridge regression
                       lambda = 0.01),  # fixed lambda
                       trControl = caret::trainControl(method = "none"))))  # Auto-uses accuracy
cat("\nMean Accuracy:", mean(cv), "\n")

## ----fig.width=7--------------------------------------------------------------
# ============================================================================
# WORKING EXAMPLES: predict_proba with unifiedml using IRIS dataset
# ============================================================================

# Load required packages
library(unifiedml)
library(randomForest)
library(nnet)
library(e1071)

# Load iris dataset
data(iris)

# Setup reproducible data
set.seed(42)

# Create feature matrix (all 4 numeric features)
X <- as.matrix(iris[, 1:4])
colnames(X) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")

# Target: Species (multi-class with 3 levels)
y_multiclass <- iris$Species

# Create binary classification target (Versicolor vs others)
y_binary <- factor(
  ifelse(iris$Species == "versicolor", "versicolor", "other"),
  levels = c("other", "versicolor")
)

# Split into train/test (75% train, 25% test)
set.seed(42)
train_idx <- sample(1:nrow(X), size = floor(0.75 * nrow(X)), replace = FALSE)
test_idx <- setdiff(1:nrow(X), train_idx)

X_train <- X[train_idx, ]
X_test <- X[test_idx, ]
y_train_multiclass <- y_multiclass[train_idx]
y_test_multiclass <- y_multiclass[test_idx]
y_train_binary <- y_binary[train_idx]
y_test_binary <- y_binary[test_idx]

cat("\n")
cat("============================================================================\n")
cat("IRIS DATASET - Summary\n")
cat("============================================================================\n")
cat(sprintf("Training samples: %d\n", nrow(X_train)))
cat(sprintf("Test samples: %d\n", nrow(X_test)))
cat(sprintf("Features: %d\n", ncol(X_train)))
cat(sprintf("Classes: %s\n", paste(levels(y_multiclass), collapse = ", ")))

# ============================================================================
# EXAMPLE 1: randomForest - Multi-class Classification on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 1: randomForest - Multi-class Classification\n")
cat("============================================================================\n")

mod_rf <- Model$new(randomForest::randomForest)
mod_rf$fit(X_train, y_train_multiclass, ntree = 100)

cat("\nPredicting probabilities for first 5 test samples:\n")
probs_rf <- mod_rf$predict_proba(X_test[1:5, ])

cat("\nProbability matrix:\n")
print(round(probs_rf, 3))

cat("\nInterpretation:\n")
for(i in 1:5) {
  cat(sprintf("\nSample %d (Actual: %s):\n", i, as.character(y_test_multiclass[i])))
  cat(sprintf("  setosa:     %.1f%%\n", probs_rf[i, "setosa"] * 100))
  cat(sprintf("  versicolor: %.1f%%\n", probs_rf[i, "versicolor"] * 100))
  cat(sprintf("  virginica:  %.1f%%\n", probs_rf[i, "virginica"] * 100))
  cat(sprintf("  Predicted:  %s\n", colnames(probs_rf)[which.max(probs_rf[i, ])]))
}

# Get class predictions
pred_classes_rf <- mod_rf$predict(X_test[1:5, ], type = "class")
cat("\nPredicted classes (first 5):", as.character(pred_classes_rf), "\n")
cat("Actual classes (first 5):   ", as.character(y_test_multiclass[1:5]), "\n")

# Calculate accuracy on full test set
probs_all_rf <- mod_rf$predict_proba(X_test)
pred_all_rf <- colnames(probs_all_rf)[apply(probs_all_rf, 1, which.max)]
accuracy_rf <- mean(pred_all_rf == as.character(y_test_multiclass))
cat(sprintf("\nTest set accuracy: %.1f%%\n", accuracy_rf * 100))

# ============================================================================
# EXAMPLE 2: nnet - Multi-class Classification on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 2: nnet - Multi-class Classification\n")
cat("============================================================================\n")

mod_nnet <- Model$new(nnet::nnet)
mod_nnet$fit(X_train, y_train_multiclass, size = 10, maxit = 200, trace = FALSE)

cat("\nPredicting probabilities for first 5 test samples:\n")
probs_nnet <- mod_nnet$predict_proba(X_test[1:5, ])

cat("\nProbability matrix (all 3 classes):\n")
print(round(probs_nnet, 3))

cat("\nDetailed predictions:\n")
for(i in 1:5) {
  cat(sprintf("\nSample %d (Actual: %s):\n", i, as.character(y_test_multiclass[i])))
  cat(sprintf("  setosa:     %.1f%%\n", probs_nnet[i, "setosa"] * 100))
  cat(sprintf("  versicolor: %.1f%%\n", probs_nnet[i, "versicolor"] * 100))
  cat(sprintf("  virginica:  %.1f%%\n", probs_nnet[i, "virginica"] * 100))
  cat(sprintf("  Predicted:  %s\n", colnames(probs_nnet)[which.max(probs_nnet[i, ])]))
}

# Get class predictions
pred_classes_nnet <- mod_nnet$predict(X_test[1:5, ], type = "class")
cat("\nPredicted classes (first 5):", as.character(pred_classes_nnet), "\n")
cat("Actual classes (first 5):   ", as.character(y_test_multiclass[1:5]), "\n")

# Calculate accuracy
probs_all_nnet <- mod_nnet$predict_proba(X_test)
pred_all_nnet <- colnames(probs_all_nnet)[apply(probs_all_nnet, 1, which.max)]
accuracy_nnet <- mean(pred_all_nnet == as.character(y_test_multiclass))
cat(sprintf("\nTest set accuracy: %.1f%%\n", accuracy_nnet * 100))

# ============================================================================
# EXAMPLE 3: SVM - Multi-class Classification on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 3: SVM - Multi-class Classification\n")
cat("============================================================================\n")

mod_svm <- Model$new(e1071::svm)
mod_svm$fit(X_train, y_train_multiclass, probability = TRUE, kernel = "radial")

cat("\nPredicting probabilities for first 5 test samples:\n")
probs_svm <- mod_svm$predict_proba(X_test[1:5, ])

cat("\nProbability matrix:\n")
print(round(probs_svm, 4))

cat("\nDetailed predictions:\n")
for(i in 1:5) {
  cat(sprintf("\nSample %d (Actual: %s):\n", i, as.character(y_test_multiclass[i])))
  cat(sprintf("  setosa:     %.1f%%\n", probs_svm[i, "setosa"] * 100))
  cat(sprintf("  versicolor: %.1f%%\n", probs_svm[i, "versicolor"] * 100))
  cat(sprintf("  virginica:  %.1f%%\n", probs_svm[i, "virginica"] * 100))
  cat(sprintf("  Predicted:  %s\n", colnames(probs_svm)[which.max(probs_svm[i, ])]))
}

# Calculate accuracy
probs_all_svm <- mod_svm$predict_proba(X_test)
pred_all_svm <- colnames(probs_all_svm)[apply(probs_all_svm, 1, which.max)]
accuracy_svm <- mean(pred_all_svm == as.character(y_test_multiclass))
cat(sprintf("\nTest set accuracy: %.1f%%\n", accuracy_svm * 100))

# ============================================================================
# EXAMPLE 4: Binary Classification on IRIS (Versicolor vs others)
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 4: Binary Classification - Versicolor vs Others\n")
cat("============================================================================\n")

# randomForest binary
mod_rf_binary <- Model$new(randomForest::randomForest)
mod_rf_binary$fit(X_train, y_train_binary, ntree = 100)

cat("\nrandomForest - Binary probabilities (first 5 test samples):\n")
probs_rf_binary <- mod_rf_binary$predict_proba(X_test[1:5, ])
print(round(probs_rf_binary, 3))

# SVM binary
mod_svm_binary <- Model$new(e1071::svm)
mod_svm_binary$fit(X_train, y_train_binary, probability = TRUE, kernel = "radial")

cat("\nSVM - Binary probabilities (first 5 test samples):\n")
probs_svm_binary <- mod_svm_binary$predict_proba(X_test[1:5, ])
print(round(probs_svm_binary, 4))

# Compare binary predictions
cat("\nComparison of Versicolor probabilities:\n")
comparison_binary <- data.frame(
  Sample = 1:5,
  Actual = as.character(y_test_binary[1:5]),
  RandomForest = round(probs_rf_binary[, "versicolor"], 3),
  SVM = round(probs_svm_binary[, "versicolor"], 4)
)
print(comparison_binary)

# ============================================================================
# EXAMPLE 5: Using unified predict() method on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 5: Using unified predict() method\n")
cat("============================================================================\n")

cat("\nrandomForest - predict(type='prob') on first 3 samples:\n")
print(round(mod_rf$predict(X_test[1:3, ], type = "prob"), 3))

cat("\nrandomForest - predict(type='class') on first 3 samples:\n")
print(mod_rf$predict(X_test[1:3, ], type = "class"))

cat("\nnnet - predict(type='class') on first 3 samples:\n")
print(mod_nnet$predict(X_test[1:3, ], type = "class"))

cat("\nSVM - predict(type='class') on first 3 samples:\n")
print(mod_svm$predict(X_test[1:3, ], type = "class"))

# ============================================================================
# EXAMPLE 6: Model Comparison on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 6: Model Performance Comparison\n")
cat("============================================================================\n")

# Compare accuracies
cat("\nModel Accuracies on IRIS test set:\n")
cat(sprintf("  randomForest: %.1f%%\n", accuracy_rf * 100))
cat(sprintf("  nnet:         %.1f%%\n", accuracy_nnet * 100))
cat(sprintf("  SVM:          %.1f%%\n", accuracy_svm * 100))

# Compare predictions for specific samples
cat("\nDetailed comparison for first 5 test samples:\n")
comparison_multi <- data.frame(
  Sample = 1:5,
  Actual = as.character(y_test_multiclass[1:5]),
  RF_Pred = as.character(mod_rf$predict(X_test[1:5, ], type = "class")),
  nnet_Pred = as.character(mod_nnet$predict(X_test[1:5, ], type = "class")),
  SVM_Pred = as.character(mod_svm$predict(X_test[1:5, ], type = "class"))
)
print(comparison_multi)

# ============================================================================
# EXAMPLE 7: Confidence Analysis on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 7: Prediction Confidence Analysis\n")
cat("============================================================================\n")

# randomForest confidence
rf_confidences <- apply(probs_all_rf, 1, max)
cat("\nrandomForest - Prediction confidence:\n")
cat(sprintf("  Mean confidence: %.1f%%\n", mean(rf_confidences) * 100))
cat(sprintf("  Median confidence: %.1f%%\n", median(rf_confidences) * 100))
cat(sprintf("  Low confidence (<70%%): %d samples (%.1f%%)\n", 
            sum(rf_confidences < 0.7), mean(rf_confidences < 0.7) * 100))
cat(sprintf("  High confidence (>90%%): %d samples (%.1f%%)\n", 
            sum(rf_confidences > 0.9), mean(rf_confidences > 0.9) * 100))

# nnet confidence
nnet_confidences <- apply(probs_all_nnet, 1, max)
cat("\nnnet - Prediction confidence:\n")
cat(sprintf("  Mean confidence: %.1f%%\n", mean(nnet_confidences) * 100))
cat(sprintf("  Median confidence: %.1f%%\n", median(nnet_confidences) * 100))
cat(sprintf("  Low confidence (<70%%): %d samples (%.1f%%)\n", 
            sum(nnet_confidences < 0.7), mean(nnet_confidences < 0.7) * 100))
cat(sprintf("  High confidence (>90%%): %d samples (%.1f%%)\n", 
            sum(nnet_confidences > 0.9), mean(nnet_confidences > 0.9) * 100))

# SVM confidence
svm_confidences <- apply(probs_all_svm, 1, max)
cat("\nSVM - Prediction confidence:\n")
cat(sprintf("  Mean confidence: %.1f%%\n", mean(svm_confidences) * 100))
cat(sprintf("  Median confidence: %.1f%%\n", median(svm_confidences) * 100))
cat(sprintf("  Low confidence (<70%%): %d samples (%.1f%%)\n", 
            sum(svm_confidences < 0.7), mean(svm_confidences < 0.7) * 100))
cat(sprintf("  High confidence (>90%%): %d samples (%.1f%%)\n", 
            sum(svm_confidences > 0.9), mean(svm_confidences > 0.9) * 100))

# ============================================================================
# EXAMPLE 8: Misclassification Analysis
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 8: Misclassification Analysis (randomForest)\n")
cat("============================================================================\n")

# Find misclassified samples
rf_misclassified <- which(pred_all_rf != as.character(y_test_multiclass))

if(length(rf_misclassified) > 0) {
  cat(sprintf("\nFound %d misclassified samples:\n", length(rf_misclassified)))
  
  for(idx in rf_misclassified[1:min(3, length(rf_misclassified))]) {
    cat(sprintf("\nSample %d:\n", idx))
    cat(sprintf("  True class: %s\n", as.character(y_test_multiclass[idx])))
    cat(sprintf("  Predicted: %s\n", pred_all_rf[idx]))
    cat("  Probabilities:\n")
    cat(sprintf("    setosa:     %.1f%%\n", probs_all_rf[idx, "setosa"] * 100))
    cat(sprintf("    versicolor: %.1f%%\n", probs_all_rf[idx, "versicolor"] * 100))
    cat(sprintf("    virginica:  %.1f%%\n", probs_all_rf[idx, "virginica"] * 100))
  }
} else {
  cat("\nPerfect classification! No misclassified samples.\n")
}

# ============================================================================
# SUMMARY
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("SUMMARY - IRIS Dataset\n")
cat("============================================================================\n")

cat("
✓ SUCCESSFUL EXAMPLES WITH IRIS DATASET:
  1. randomForest - Multi-class classification (3 species)
  2. nnet - Multi-class classification
  3. SVM - Multi-class classification with probabilities
  4. Binary classification (Versicolor vs others)
  5. Unified predict() interface
  6. Model comparison and accuracy analysis
  7. Confidence analysis
  8. Misclassification analysis

✓ KEY FINDINGS ON IRIS:
  • All models achieve high accuracy (>90%) on iris dataset
  • SVM tends to produce extreme probabilities (near 0 or 1)
  • randomForest and nnet show more calibrated probabilities
  • Setosa is perfectly separable from other species
  • Confusion typically occurs between versicolor and virginica

✓ predict_proba() FEATURES DEMONSTRATED:
  • Returns matrix [n_samples × 3] for multi-class
  • Column names: setosa, versicolor, virginica
  • All rows sum to 1
  • Works seamlessly across all model types

All working examples on IRIS dataset completed successfully!\n")

## -----------------------------------------------------------------------------
library(unifiedml)

## -----------------------------------------------------------------------------
# Example 1: lm with matrix interface (including factors)
demo_lm_matrix <- function() {
  cat("\n=== Example 1: lm with matrix interface ===\n")
  
  lm_matrix <- unifiedml::formula_to_matrix(stats::lm)  # Uses generic predict()
  
  # Create data with factor and numeric columns
  X <- data.frame(
    wt = mtcars$wt,
    hp = mtcars$hp,
    cyl = factor(mtcars$cyl)
  )
  y <- mtcars$mpg
  weights <- rep(1, nrow(X))
  
  model <- lm_matrix$fit(X, y, weights = weights)
  preds <- lm_matrix$predict(model, X[1:5, ])
  
  cat("Predictions:\n")
  print(preds)
  cat("\nCoefficients:\n")
  print(coef(model))
}

# Example 2: glmnet with formula interface (with factors)
demo_glmnet_formula <- function() {
  cat("\n=== Example 2: glmnet with formula interface ===\n")
  
  if (!requireNamespace("glmnet", quietly = TRUE)) {
    cat("glmnet package not installed, skipping example\n")
    return(invisible(NULL))
  }
  
  glmnet_formula <- unifiedml::matrix_to_formula(
    fit_func = glmnet::glmnet,
    predict_func = function(model, newX, ...) {
      # Wrapper to provide default s parameter
      glmnet::predict.glmnet(model, newx = newX, s = 0.01, ...)
    }
  )
  
  # Formula with factor - model.matrix will auto-create dummies
  model <- glmnet_formula$fit(mpg ~ wt + hp + factor(cyl), data = mtcars)
  preds <- glmnet_formula$predict(model, newdata = mtcars[1:5, ])
  
  cat("Predictions:\n")
  print(preds)
}

# Example 3: Special characters in column names
demo_special_names <- function() {
  cat("\n=== Example 3: Column names with special characters ===\n")
  
  lm_matrix <- unifiedml::formula_to_matrix(lm)
  
  # Create problematic column names
  X <- data.frame(
    `x-1` = mtcars$wt,      # Dash
    `a:b` = mtcars$hp,       # Colon
    `log(x)` = log(mtcars$disp),  # Parentheses
    check.names = FALSE
  )
  y <- mtcars$mpg
  
  model <- lm_matrix$fit(X, y)
  preds <- lm_matrix$predict(model, X[1:3, ])
  
  cat("Predictions:\n")
  print(preds)
  cat("\nModel handled special column names correctly!\n")
}

# Run examples (uncomment to test)
demo_lm_matrix()
demo_glmnet_formula()
demo_special_names()

#' Cross-Validation for Model Objects
#'
#' Perform k-fold cross-validation with consistent scoring metrics
#' across different model types. The scoring metric is automatically
#' selected based on the detected task type.
#'
#' @param model A Model object
#' @param X Feature matrix or data.frame
#' @param y Target vector (type determines regression vs classification)
#' @param cv Number of cross-validation folds (default: 5)
#' @param scoring Scoring metric: "rmse", "mae", "accuracy", "f1", or a
#'   custom function with signature \code{function(true, pred)} returning
#'   a scalar. Default: auto-detected based on task type.
#' @param show_progress Whether to show progress bar (default: TRUE) in sequential mode
#' @param verbose logical flag enabling verbose messages (default: TRUE) in parallel mode
#' @param cl Optional number of clusters for parallel processing
#' If using \code{cl} for parallel execution, custom scoring functions must
#' be self-contained (no dependencies on the calling environment).
#' @param seed Reproducibility seed
#' @param fit_params A list of additional arguments passed to model$fit()  
#' @param predict_params A list of additional arguments passed to model$predict()
#'
#' @return Vector of cross-validation scores for each fold
#'
#' @examples
#' \dontrun{
#' library(glmnet)
#' X <- matrix(rnorm(100), ncol = 4)
#' y <- 2*X[,1] - 1.5*X[,2] + rnorm(25)  # numeric -> regression
#'
#' mod <- Model$new(glmnet::glmnet)
#' (cv_scores <- cross_val_score(mod, X, y, cv = 5))  # auto-uses RMSE
#' mean(cv_scores)  # Average RMSE
#'
#' cross_val_score(mod, X, y,
#' fit_params     = list(alpha = 0, lambda = 0.1),
#' predict_params = list(type = "response"))
#' 
#' cross_val_score(mod, X, y,
#' fit_params     = list(alpha = 0.5, lambda = 0.1),
#' predict_params = list(type = "response"))
#' 
#' # Custom scoring: R-squared
#' r2 <- function(true, pred) {
#'   ss_res <- sum((true - pred)^2)
#'   ss_tot <- sum((true - mean(true))^2)
#'   1 - ss_res / ss_tot
#' }
#' 
#' (cv_scores4 <- cross_val_score(mod, X, y, cv = 5, scoring = r2))
#' mean(cv_scores4)  # Average R²
#' 
#' # Classification with accuracy scoring
#' data(iris)
#' X_class <- iris[, 1:4]
#' y_class <- iris$Species  # factor -> classification
#' mod2 <- Model$new(e1071::svm)
#' (cv_scores2 <- cross_val_score(mod2, X_class, y_class, cv = 5))  # auto-uses accuracy
#' mean(cv_scores2)  # Average accuracy
#' 
#' iris_bin <- iris[iris$Species != "virginica", ]
#' X_bin <- iris_bin[, 1:4]
#' y_bin <- droplevels(iris_bin$Species)
#' (cv_scores3 <- cross_val_score(mod2, X_bin, y_bin, cv = 3, 
#' scoring="f1", fit_params=list(kernel="polynomial")))  
#' mean(cv_scores3)  # Average F1
#' }
#'
#' @export
cross_val_score <- function(model,
                            X,
                            y,
                            cv = 5,
                            scoring = NULL,
                            show_progress = TRUE,
                            verbose = TRUE,
                            cl = NULL,
                            seed = 123,
                            fit_params = NULL, 
                            predict_params = NULL) {
  #X <- as.matrix(X)
  n <- nrow(X)
  if (length(y) != n)
    stop("Must have: length(y) == nrow(X)")
  if (!(floor(cv) == cv))
    stop("'cv' must be an integer")
  
  set.seed(seed)
  
  folds <- base::split(base::sample(seq_len(n)), rep(1:cv, length.out = n))
  scores <- numeric(cv)
  names(scores) <- paste0("fold", seq_len(cv))
  
  # Auto-detect task based on y
  task_type <- ifelse(is.factor(y), "classification", "regression")
  
  # Auto-detect scoring metric if not provided
  if (is.null(scoring)) {
    scoring <- ifelse(task_type == "regression", "rmse", "accuracy")
  }
  
  # resolve scoring to a function once, before the fold loop
  score_fn <- if (is.function(scoring)) {
    scoring
  } else {
    switch(scoring,
           rmse     = function(true, pred) sqrt(mean((true - pred)^2, na.rm = TRUE)),
           mae      = function(true, pred) mean(abs(true - pred), na.rm = TRUE),
           accuracy = function(true, pred) mean(pred == true, na.rm = TRUE),
           f1       = function(true, pred) compute_f1(true, pred), # see implementation below
           stop("Unknown scoring metric: ", scoring)
    )
  }
  
  if (is.null(cl))
  {
    if (show_progress)
      pb <- utils::txtProgressBar(max = cv, style = 3)
    
    for (i in seq_len(cv)) {
      val_idx   <- folds[[i]]
      train_idx <- setdiff(seq_len(n), val_idx)
      
      m <- model$clone_model()
      do.call(m$fit, c(list(X = X[train_idx,, drop=FALSE], y = y[train_idx]), fit_params))
      
      pred <- do.call(m$predict, c(list(X = X[val_idx,,  drop=FALSE]), predict_params))
      if (is.factor(y))
      {
        true <- droplevels(y[val_idx])
      } else {
        true <- y[val_idx]
      }
      
      scores[i] <- score_fn(true, pred)
      
      if (show_progress)
        utils::setTxtProgressBar(pb, i)
    }
    
    if (show_progress)
      close(pb)
    
    return(scores)
    
  } else {
    cl_SOCK <- parallel::makeCluster(cl, type = "SOCK")
    doParallel::registerDoParallel(cl_SOCK)
    `%op%` <-  foreach::`%dopar%`
    
    # KEY FIX: Store the result and use .combine to collect scores
    scores <- foreach::foreach(
      i = seq_len(cv),
      .packages = c("unifiedml"),
      .combine = 'c',
      .errorhandling = "stop",
      .verbose = verbose
    ) %op% {
      val_idx   <- folds[[i]]
      train_idx <- setdiff(seq_len(n), val_idx)
      
      m <- model$clone_model()
      do.call(m$fit,    c(list(X = X[train_idx,, drop=FALSE], y = y[train_idx]), fit_params))
      
      pred <- do.call(m$predict, c(list(X = X[val_idx,,  drop=FALSE]), predict_params))
      if (is.factor(y))
      {
        true <- droplevels(y[val_idx])
      } else {
        true <- y[val_idx]
      }
      
      # RETURN the score (not assign to scores[i])
      score_fn(true, pred)
    }
    
    parallel::stopCluster(cl_SOCK)
    names(scores) <- paste0("fold", seq_len(cv))
    return(scores)
  }
  
}

compute_f1 <- function(true, pred)
{
  if (is.factor(true)) {
    # For binary classification, assume first two levels
    if (nlevels(true) == 2) {
      tp <- sum(pred == levels(true)[2] & true == levels(true)[2])
      fp <- sum(pred == levels(true)[2] &
                  true == levels(true)[1])
      fn <- sum(pred == levels(true)[1] &
                  true == levels(true)[2])
      precision <- tp / (tp + fp + 1e-10)
      recall <- tp / (tp + fn + 1e-10)
      score <- 2 * precision * recall / (precision + recall + 1e-10)
    } else {
      warning(
        "F1 score currently only supports binary classification. Using accuracy instead."
      )
      score <- mean(pred == true, na.rm = TRUE)
    }
  } else {
    # For numeric binary classification (0/1)
    tp <- sum(pred == 1 &
                true == 1)
    fp <- sum(pred == 1 &
                true == 0)
    fn <- sum(pred == 0 &
                true == 1)
    precision <- tp / (tp + fp + 1e-10)
    recall <- tp / (tp + fn + 1e-10)
    score <- 2 * precision * recall / (precision + recall + 1e-10)
  }
  return(score)
}

#' Split data into training and test sets
#'
#' Randomly splits a feature matrix or data.frame and its corresponding
#' response vector into training and test subsets.
#'
#' @param X A matrix or data.frame of features.
#' @param y A vector of responses (numeric or factor). Must have the same
#'   number of rows as \code{X}.
#' @param test_size Proportion of observations to use as the test set.
#'   A number in (0, 1). Default is \code{0.2} (80/20 split).
#' @param seed An optional integer random seed for reproducibility. If
#'   \code{NULL} (default) the current RNG state is used.
#'
#' @return A named list with four elements:
#'   \item{X_train}{Training features (same type as \code{X}).}
#'   \item{X_test}{Test features (same type as \code{X}).}
#'   \item{y_train}{Training response.}
#'   \item{y_test}{Test response.}
#'
#' @examples
#' # matrix input
#' X <- iris[, 1:4]
#' y <- iris$Species
#' d <- unifiedml::train_test_split(X, y, test_size = 0.3, seed = 42)
#' dim(d$X_train)  # 105 x 4
#' dim(d$X_test)   #  45 x 4
#'
#' # data.frame input
#' d2 <- unifiedml::train_test_split(iris[, 1:4], iris$Species, test_size = 0.2)
#' is.data.frame(d2$X_train)  # TRUE
#'
#' @export
train_test_split <- function(X, y, test_size = 0.2, seed = NULL) {
  stopifnot(
    is.matrix(X) || is.data.frame(X),
    is.vector(y) || is.factor(y),
    nrow(X) == length(y),
    test_size > 0 && test_size < 1
  )
  
  if (!is.null(seed)) set.seed(seed)
  
  n        <- nrow(X)
  n_test   <- max(1L, floor(n * test_size))
  test_idx <- sample(n, size = n_test)
  
  list(
    X_train = X[-test_idx, , drop = FALSE],
    X_test  = X[ test_idx, , drop = FALSE],
    y_train = y[-test_idx],
    y_test  = y[ test_idx]
  )
}

#' Unified Machine Learning Interface using R6
#' 
#' Provides a consistent interface for various machine learning models in R,
#' with automatic detection of formula vs matrix interfaces, built-in
#' cross-validation, model interpretability, and visualization.
#' 
#' @name Model
#' @docType class
#' @author Your Name
#' 
#' @import R6
#' @importFrom stats predict pt sd cor
#' @importFrom utils txtProgressBar setTxtProgressBar
NULL

#' @title Unified Machine Learning Model
#' @description
#' An R6 class that provides a unified interface for regression and classification
#' models with automatic interface detection, cross-validation, and interpretability
#' features. The task type (regression vs classification) is automatically detected
#' from the response variable type.
#' 
#' @field model_fn The modeling function (e.g., glmnet::glmnet, randomForest::randomForest)
#' @field fitted The fitted model object
#' @field task Type of task: "regression" or "classification" (automatically detected)
#' @field X_train Training features matrix
#' @field y_train Training target vector
#' 
#' @examples
#' \donttest{
#' # Regression example with glmnet
#' library(glmnet)
#' X <- matrix(rnorm(100), ncol = 4)
#' y <- 2*X[,1] - 1.5*X[,2] + rnorm(25)  # numeric -> regression
#' 
#' mod <- Model$new(glmnet::glmnet)
#' mod$fit(X, y, alpha = 0, lambda = 0.1)
#' mod$summary()
#' predictions <- mod$predict(X)
#' 
#' # Classification example  
#' data(iris)
#' iris_binary <- iris[iris$Species %in% c("setosa", "versicolor"), ]
#' X_class <- as.matrix(iris_binary[, 1:4])
#' y_class <- droplevels(iris_binary$Species)  # factor -> classification
#' 
#' mod2 <- Model$new(e1071::svm)
#' mod2$fit(X_class, y_class, kernel = "radial")
#' predictions <- mod2$predict(X_class)
#' mod2$predict_proba(X_class)
#' 
#' }
#' 
#' @export
Model <- R6::R6Class(
  "Model",
  public = list(
    model_fn = NULL,
    fitted   = NULL,
    task     = NULL,  # Will be set automatically in fit()
    X_train  = NULL,
    y_train  = NULL,
    
    #' @description Initialize a new Model
    #' @param model_fn A modeling function (e.g., glmnet, randomForest, svm)
    #' @return A new Model object
    initialize = function(model_fn) {
      stopifnot(is.function(model_fn))
      self$model_fn <- model_fn
    },
    
    #' @description Fit the model to training data
    #' 
    #' Automatically detects task type (regression vs classification) based on
    #' the type of the response variable y. Numeric y -> regression, 
    #' factor y -> classification.
    #' 
    #' @param X Feature matrix or data.frame
    #' @param y Target vector (numeric for regression, factor for classification)
    #' @param ... Additional arguments passed to the model function
    #' @return self (invisible) for method chaining
    fit = function(X, y, ...) {
      X <- as.matrix(X)
      
      # Store training data for later use
      self$X_train <- X
      self$y_train <- y
      
      # Auto-detect task type based on y
      if (is.factor(y)) {
        self$task <- "classification"
        # Ensure y is factor for classification
        y <- as.factor(y)
      } else {
        self$task <- "regression"
        # Ensure y is numeric for regression
        y <- as.numeric(y)
      }
      
      # 1. Try formula + data interface
      df <- data.frame(y = y, X)
      fit_args <- c(list(formula = y ~ ., data = df), list(...))
      self$fitted <- tryCatch(
        do.call(self$model_fn, fit_args),
        error = function(e) NULL
      )
      
      # 2. If failed -> try matrix interface
      if (is.null(self$fitted)) {
        fit_args <- c(list(x = X, y = y), list(...))
        self$fitted <- tryCatch(
          do.call(self$model_fn, fit_args),
          error = function(e) {
            fit_args <- c(list(x = as.matrix(X), y = as.integer(y)), list(...))
            self$fitted <- tryCatch(do.call(self$model_fn, fit_args),
                                    error = function(e) {
                                      print(e)
                                      stop("Model fit failed.")
                                    }
            )
          }
        )
      }
      
      invisible(self)
    },
    
    #' @description Generate predictions from fitted model
    #' @param X Feature matrix for prediction
    #' @param ... Additional arguments passed to predict function
    #' @return Vector of predictions
    predict = function(X, ...) {
      if (is.null(self$fitted)) stop("Model not fitted.")
      X <- as.matrix(X)
      
      # 1. Try newdata (formula models)
      df <- data.frame(X)
      pred <- tryCatch(
        predict(self$fitted, df, ...),
        error = function(e) NULL
      )
      
      # 2. Fallback: newx (matrix models)
      if (is.null(pred)) {
        pred <- tryCatch(
          predict(self$fitted, X, ...),
          error = function(e) {
            stop("Predict failed.")
          }
        )
      }
      
      if (is.list(pred)) # example: ranger
        return(pred)
      
      if (!is.null(dim(pred)) && (dim(pred)[2] != 1) && is.numeric(pred)) # probabilities array
      {
        return(pred)
      }
      
      # Clean output
      if (is.matrix(pred) && ncol(pred) == 1) pred <- drop(pred)
      
      # For classification, ensure factors are returned as original levels if possible
      if (self$task == "classification" && is.factor(self$y_train)) {
        if (is.numeric(pred)) {
          # Convert numeric predictions back to factor levels
          pred <- factor(levels(self$y_train)[pred + 1], levels = levels(self$y_train))
        } else {
          pred <- droplevels(as.factor(pred))
        }
      }
      
      drop(pred)
    },
    
    
    #' @description Predict probabilities from fitted model
    #' @param X Feature matrix for prediction
    predict_proba = function(X) {
      if (is.null(self$fitted)) stop("Model not fitted.")
      if (self$task != "classification") {
        stop("predict_proba() only for classification tasks")
      }
      extract_probabilities(self$fitted, X, self$y_train)
    },
    
    #' @description Print model information
    #' @return self (invisible) for method chaining
    print = function() {
      cat("Model Object\n")
      cat("------------\n")
      cat("Model function:", deparse(substitute(self$model_fn))[1], "\n")
      cat("Fitted:", !is.null(self$fitted), "\n")
      if (!is.null(self$fitted)) {
        cat("Task:", self$task, "\n")
        cat("Training samples:", nrow(self$X_train), "\n")
        cat("Features:", ncol(self$X_train), "\n")
        if (self$task == "classification") {
          cat("Classes:", paste(levels(self$y_train), collapse = ", "), "\n")
          cat("Class distribution:\n")
          print(table(self$y_train))
        }
      }
      invisible(self)
    },
    
    #' @description Compute numerical derivatives and statistical significance
    #' 
    #' Uses finite differences to compute approximate partial derivatives
    #' for each feature, providing model-agnostic interpretability.
    #' 
    #' @param h Step size for finite differences (default: 0.01)
    #' @param alpha Significance level for p-values (default: 0.05)
    #' @return A data.frame with derivative statistics (invisible)
    #' 
    #' @details
    #' The method computes numerical derivatives using central differences.
    #' 
    #' Statistical significance is assessed using t-tests on the derivative
    #' estimates across samples.
    summary = function(h = 0.01, alpha = 0.05) {
      if (is.null(self$fitted)) stop("Model not fitted.")
      
      n <- nrow(self$X_train)
      p <- ncol(self$X_train)
      
      # Compute numerical derivatives for each feature
      derivatives <- matrix(0, nrow = n, ncol = p)
      
      for (j in 1:p) {
        X_plus <- X_minus <- self$X_train
        X_plus[, j] <- X_plus[, j] + h
        X_minus[, j] <- X_minus[, j] - h
        
        pred_plus <- self$predict(X_plus)
        pred_minus <- self$predict(X_minus)
        
        derivatives[, j] <- (pred_plus - pred_minus) / (2 * h)
      }
      
      # Compute mean derivatives and standard errors
      mean_deriv <- colMeans(derivatives)
      se_deriv <- apply(derivatives, 2, sd) / sqrt(n)
      
      # Compute t-statistics and p-values
      t_stat <- mean_deriv / se_deriv
      p_values <- 2 * pt(-abs(t_stat), df = n - 1)
      
      # Create summary table
      feature_names <- colnames(self$X_train)
      if (is.null(feature_names)) {
        feature_names <- paste0("X", 1:p)
      }
      
      summary_table <- data.frame(
        Feature = feature_names,
        Mean_Derivative = mean_deriv,
        Std_Error = se_deriv,
        t_value = t_stat,
        p_value = p_values,
        Significance = ifelse(p_values < alpha, "***", 
                              ifelse(p_values < alpha * 2, "**",
                                     ifelse(p_values < alpha * 4, "*", "")))
      )
      
      cat("\nModel Summary - Numerical Derivatives\n")
      cat("======================================\n")
      cat("Task:", self$task, "\n")
      cat("Samples:", n, "| Features:", p, "\n")
      cat("Step size (h):", h, "\n\n")
      
      print(summary_table, row.names = FALSE)
      cat("\nSignificance codes: 0 '***' 0.01 '**' 0.05 '*' 0.1 ' ' 1\n")
      
      invisible(summary_table)
    },
    
    #' @description Create partial dependence plot for a feature
    #' 
    #' Visualizes the relationship between a feature and the predicted outcome
    #' while holding other features at their mean values.
    #' 
    #' @param feature Index or name of feature to plot
    #' @param n_points Number of points for the grid (default: 100)
    #' @return self (invisible) for method chaining
    plot = function(feature = 1, n_points = 100) {
      if (is.null(self$fitted)) stop("Model not fitted.")
      
      p <- ncol(self$X_train)
      if (feature < 1 || feature > p) {
        stop("feature must be between 1 and ", p)
      }
      
      feature_name <- colnames(self$X_train)[feature]
      if (is.null(feature_name)) feature_name <- paste0("X", feature)
      
      # Create a grid for the selected feature
      x_range <- range(self$X_train[, feature])
      x_grid <- seq(x_range[1], x_range[2], length.out = n_points)
      
      # Hold other features at their means
      X_grid <- matrix(rep(colMeans(self$X_train), each = n_points), 
                       nrow = n_points)
      X_grid[, feature] <- x_grid
      
      # Get predictions
      y_pred <- self$predict(X_grid)
      
      # Create plot
      #par(mfrow = c(1, 1))
      plot(self$X_train[, feature], self$y_train,
           xlab = feature_name,
           ylab = ifelse(self$task == "regression", "y", "Class"),
           main = paste("Partial Dependence Plot -", feature_name),
           pch = 16, col = rgb(0, 0, 0, 0.3))
      
      lines(x_grid, y_pred, col = "red", lwd = 2)
      
      invisible(self)
    },
    
    #' @description Create a deep copy of the model
    #' 
    #' Useful for cross-validation and parallel processing where
    #' multiple independent model instances are needed.
    #' 
    #' @return A new Model object with same configuration
    clone_model = function() {
      Model$new(self$model_fn)
    }
  )
)

#' Benchmark Multiple Models with Cross-Validation and Model-Specific Parameters
#'
#' Perform k-fold cross-validation on a list of models, using model-specific parameters.
#' Supports verbose messages and a progress bar.
#'
#' @param models A named list of \code{Model$new(...)} objects to benchmark.
#' @param X A data frame or matrix of predictors.
#' @param y A vector of outcomes (factor for classification, numeric for regression).
#' @param cv Integer, number of cross-validation folds (default 5).
#' @param scoring Scoring metric: "rmse", "mae", "accuracy", or "f1" 
#'               (default: auto-detected based on task)
#' @param params Optional named list of lists, each sublist containing extra arguments
#'   to pass to the corresponding model's \code{fit()} call. Names must match `models`.
#' @param cl Optional number of clusters for parallel processing
#' @param show_progress Logical, whether to show a progress bar (default TRUE).
#' @param verbose Logical, whether to print messages about each model (default TRUE).
#'
#' @return A list containing the CV scores for each model.
#'
#' @examples
#' \dontrun{
#' library(randomForest)
#'
#' X <- iris[, 1:4]
#' y <- iris$Species
#'
#' models <- list(
#'   glm  = Model$new(caret::train),
#'   rf   = Model$new(randomForest::randomForest),
#'   xgb  = Model$new(caret::train)
#' )
#'
#' params <- list(
#'   glm = list(method = "glmnet",
#'              tuneGrid = data.frame(alpha = 0, lambda = 0.01),
#'              trControl = trainControl(method = "none")),
#'   rf  = list(ntree = 150),
#'   xgb = list(method = "xgbTree",
#'              tuneGrid = data.frame(nrounds = 150, max_depth = 3, eta = 0.3,
#'                                    gamma = 0, colsample_bytree = 1,
#'                                    min_child_weight = 1, subsample = 1),
#'              trControl = trainControl(method = "none"))
#' )
#'
#' results <- benchmark(models, X, y, cv = 5, params = params,
#'                      show_progress = TRUE, verbose = TRUE)
#' print(results)
#' }
#' @export
benchmark <- function(models, X, y, cv = 5L, scoring = NULL, params = NULL, cl=NULL, show_progress = FALSE, verbose = TRUE) {
  n_models <- length(models)
  results <- vector("list", n_models)
  names(results) <- names(models)
  
  if (show_progress) {
    pb <- utils::txtProgressBar(min = 0, max = n_models, style = 3)
  }
  
  for (i in seq_along(models)) {
    model_name <- names(models)[i]
    mod <- models[[i]]
    
    if (verbose) cat(sprintf("\n[%d/%d] Fitting model: %s\n", 
                             i, length(models), model_name))
    
    # Extract model-specific parameters if provided
    extra_args <- if (!is.null(params) && model_name %in% names(params)) {
      params[[model_name]]
    } else {
      list()
    }
    
    scores <- cross_val_score(
      model = mod,
      X = X,
      y = y,
      cv = cv,
      scoring = scoring,
      show_progress = FALSE,
      cl = cl,
      fit_params = extra_args
    )
    
    results[[i]] <- list(avg_score = mean(scores), scores = scores)
    
    if (verbose) cat(sprintf("Mean CV score for %s: %.4f\n", model_name, results[[i]]$avg_score))
    
    if (show_progress) utils::setTxtProgressBar(pb, i)
  }
  
  if (show_progress) close(pb)
  
  return(results)
}

#' @importFrom stats predict family model.matrix
NULL

# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------

.get_class_labels <- function(y) {
  if (is.factor(y))          return(levels(y))
  if (is.character(y))       return(sort(unique(y)))
  if (is.numeric(y))         return(c("0", "1"))
  NULL
}

.to_matrix <- function(X) {
  if (!is.matrix(X)) as.matrix(X) else X
}

.to_df <- function(X) {
  if (!is.data.frame(X)) as.data.frame(X) else X
}

# Dispatcher table: each entry is function(model, X) -> raw prediction or NULL
.extractors <- list(
  
  randomForest = function(model, X) {
    if (model$type != "classification") return(NULL)
    tryCatch(predict(model, X, type = "prob"), error = function(e) NULL)
  },
  
  glm = function(model, X) {
    fam <- tryCatch(family(model)$family, error = function(e) NULL)
    if (is.null(fam)) return(NULL)
    X_df <- .to_df(X)
    if (fam == "binomial") {
      tryCatch(predict(model, newdata = X_df, type = "response"), error = function(e) NULL)
    } else {
      NULL  # multinomial is not a base-R glm family; handled via fallback
    }
  },
  
  multinom = function(model, X) {
    tryCatch(predict(model, newdata = .to_df(X), type = "probs"), error = function(e) NULL)
  },
  
  svm = function(model, X) {
    if (model$type != 0L) return(NULL)  # 0 = C-classification
    tryCatch({
      pred_obj <- predict(model, X, probability = TRUE)
      attr(pred_obj, "probabilities")
    }, error = function(e) NULL)
  },
  
  nnet = function(model, X) {
    tryCatch(predict(model, newdata = .to_df(X), type = "raw"), error = function(e) NULL)
  },
  
  rpart = function(model, X) {
    tryCatch(predict(model, newdata = .to_df(X), type = "prob"), error = function(e) NULL)
  },
  
  xgb.Booster = function(model, X) {
    tryCatch({
      pred <- predict(.to_matrix(X))
      if (is.null(dim(pred))) cbind(1 - pred, pred) else pred
    }, error = function(e) NULL)
  },
  
  glmnet = function(model, X) {
    tryCatch({
      pred <- predict(model, .to_matrix(X), type = "response")
      if (is.matrix(pred) && ncol(pred) == 1) cbind(1 - pred, pred) else pred
    }, error = function(e) NULL)
  },
  
  ksvm = function(model, X) {
    tryCatch(
      predict(model, .to_matrix(X), type = "probabilities"),
      error = function(e) NULL
    )
  }
)

# Generic fallback cascade - tried in order when no specific extractor matches
.fallback_extractors <- list(
  
  function(model, X)
    tryCatch(predict(model, newdata = .to_df(X),  type = "response"),   error = function(e) NULL),
  
  function(model, X)
    tryCatch(predict(model, newx = .to_matrix(X), type = "response"),   error = function(e) NULL),
  
  function(model, X)
    tryCatch(predict(model, X,                    type = "response"),   error = function(e) NULL),
  
  function(model, X)
    tryCatch(predict(model, newdata = .to_df(X),  type = "prob"),       error = function(e) NULL),
  
  function(model, X)
    tryCatch(predict(model, newdata = .to_df(X),  type = "raw"),        error = function(e) NULL),
  
  function(model, X)
    tryCatch({
      pred_obj <- predict(model, X, probability = TRUE)
      probs    <- attr(pred_obj, "probabilities")
      if (is.matrix(probs)) probs else NULL
    }, error = function(e) NULL),
  
  function(model, X)
    tryCatch(predict(model, X), error = function(e) NULL)
)

# ---------------------------------------------------------------------------
# Standardise raw prediction -> n x k probability matrix
# ---------------------------------------------------------------------------

.standardize_to_probs <- function(pred, y_train, verbose) {
  
  # Case 1: already a valid probability matrix
  if (is.matrix(pred) && all(pred >= 0) && all(pred <= 1)) {
    row_sums <- rowSums(pred)
    if (all(abs(row_sums - 1) < 1e-6)) {
      if (verbose) message("  Already a probability matrix")
      if (is.null(colnames(pred)) && !is.null(y_train))
        colnames(pred) <- .get_class_labels(y_train)
      return(pred)
    }
  }
  
  # Case 2: numeric vector
  if (is.numeric(pred) && !is.matrix(pred)) {
    if (all(pred >= 0 & pred <= 1)) {
      if (verbose) message("  Binary probability vector")
      n_classes <- if (!is.null(y_train) && is.factor(y_train)) nlevels(y_train) else 2
      if (n_classes != 2)
        stop("Probability vector implies binary classification but y_train has ", n_classes, " classes")
      prob_matrix <- cbind(1 - pred, pred)
      colnames(prob_matrix) <- if (!is.null(y_train))
        .get_class_labels(y_train) else c("Class0", "Class1")
      return(prob_matrix)
    } else {
      warning("Numeric predictions outside [0,1] interpreted as logits; verify this is correct")
      if (verbose) message("  Logit vector -> sigmoid")
      probs <- 1 / (1 + exp(-pred))
      return(.standardize_to_probs(probs, y_train, verbose))
    }
  }
  
  # Case 3: factor (hard classes) -> one-hot
  if (is.factor(pred)) {
    if (verbose) message("  Hard classes -> one-hot")
    classes      <- levels(pred)
    prob_matrix  <- model.matrix(~ pred - 1)
    colnames(prob_matrix) <- classes
    return(prob_matrix)
  }
  
  # Case 4: matrix needing normalisation
  if (is.matrix(pred)) {
    row_sums <- rowSums(pred)
    if (any(abs(row_sums - 1) > 1e-6)) {
      if (all(pred >= 0)) {
        if (verbose) message("  Raw score matrix -> row normalise")
        prob_matrix <- pred / row_sums
      } else {
        if (verbose) message("  Logit matrix -> softmax")
        exp_pred    <- exp(pred - apply(pred, 1, max))  # numerically stable
        prob_matrix <- exp_pred / rowSums(exp_pred)
      }
      if (!is.null(y_train))
        colnames(prob_matrix) <- .get_class_labels(y_train)
      return(prob_matrix)
    }
  }
  
  # Case 5: list with $probabilities
  if (is.list(pred) && !is.null(pred$probabilities)) {
    if (verbose) message("  List$probabilities")
    return(.standardize_to_probs(pred$probabilities, y_train, verbose))
  }
  
  # Last resort
  if (verbose) message("  Unknown format, coercing to matrix")
  .standardize_to_probs(as.matrix(pred), y_train, verbose)
}

# ---------------------------------------------------------------------------
# Public function
# ---------------------------------------------------------------------------

#' Extract probability predictions from any R model in a standardised format
#'
#' @param model Fitted model object (any class)
#' @param X Feature matrix or data.frame for predictions
#' @param y_train Optional training labels used to name output columns
#' @param verbose Print diagnostic information (default: FALSE)
#'
#' @return Numeric matrix of shape n_samples x n_classes with column names
#'   as class labels. Attributes \code{extraction_method} and
#'   \code{model_class} record how predictions were obtained.
#'
#' @export
extract_probabilities <- function(model, X, y_train = NULL, verbose = FALSE) {
  
  # Input validation
  if (is.null(model))           stop("'model' must not be NULL")
  if (NROW(X) == 0)             stop("'X' has no rows")
  if (!is.null(y_train) && !is.factor(y_train) &&
      !is.character(y_train) && !is.numeric(y_train))
    stop("'y_train' must be a factor, character, or numeric vector")
  
  model_class <- class(model)[1]
  if (verbose) message("Model class: ", paste(class(model), collapse = ", "))
  
  # 1. Try registered extractor
  extractor <- .extractors[[model_class]]
  prediction <- if (!is.null(extractor)) {
    if (verbose) message("  Trying registered extractor for '", model_class, "'")
    extractor(model, X)
  } else NULL
  
  method <- if (!is.null(prediction))
    paste0("registered::", model_class) else "unknown"
  
  # 2. Fall back to generic cascade
  if (is.null(prediction)) {
    if (verbose) message("  No registered extractor, trying fallback cascade...")
    for (i in seq_along(.fallback_extractors)) {
      prediction <- .fallback_extractors[[i]](model, X)
      if (!is.null(prediction)) {
        method <- paste0("fallback::", i)
        if (verbose) message("  Fallback ", i, " succeeded")
        break
      }
    }
  }
  
  if (is.null(prediction))
    stop("Could not extract predictions from model of class: ",
         paste(class(model), collapse = ", "))
  
  if (verbose) message("Extraction method: ", method)
  
  # Standardise to probability matrix
  result <- .standardize_to_probs(prediction, y_train, verbose)
  
  # Final checks
  if (!is.matrix(result))
    result <- as.matrix(result)
  
  result[result < 0] <- 0
  result[result > 1] <- 1
  
  row_sums <- rowSums(result)
  if (any(abs(row_sums - 1) > 0.01))
    warning("Row sums deviate from 1 by more than 0.01 before normalisation; check model output")
  if (any(abs(row_sums - 1) > 1e-4))
    result <- result / row_sums
  
  attr(result, "extraction_method") <- method
  attr(result, "model_class")       <- model_class
  
  if (verbose) {
    message("Output: ", paste(dim(result), collapse = "x"), " matrix")
    message("Classes: ", paste(colnames(result), collapse = ", "))
  }
  
  result
}

## ----fig.width=7, eval=TRUE---------------------------------------------------
library(unifiedml)
library(randomForest)
library(e1071)
library(caret)
set.seed(123)

X <- iris[, 1:4]
y <- iris$Species

models <- list(
  glm  = Model$new(caret::train),
  rf   = Model$new(randomForest::randomForest),
  svm  = Model$new(e1071::svm)   
)

params <- list(
  glm = list(method = "glmnet",
             tuneGrid = data.frame(alpha = 0, lambda = 0.01),
             trControl = trainControl(method = "none")),
  
  rf  = list(ntree = 150),
  
  svm = list(kernel = "radial",   # <-- added
             cost = 1,
             gamma = 0.1)
)

results <- benchmark(models, X, y, cv = 5, params = params)
print(results)

## ----eval=TRUE----------------------------------------------------------------
library(unifiedml)
library(randomForest)
library(e1071)
library(caret)

set.seed(123)

# Regression data
X <- mtcars[, setdiff(names(mtcars), "mpg")]
y <- mtcars$mpg

models <- list(
  glm  = Model$new(caret::train),
  rf   = Model$new(randomForest::randomForest),
  svm  = Model$new(e1071::svm)
)

params <- list(
  glm = list(method = "glmnet",
             tuneGrid = data.frame(alpha = 0, lambda = 0.01),
             trControl = trainControl(method = "none")),
  
  rf  = list(ntree = 150),
  
  svm = list(type = "eps-regression",  # <-- important for regression
             kernel = "radial",
             cost = 1,
             gamma = 0.1)
)

results <- benchmark(models, X, y, cv = 5, params = params)
print(results)

#' Convert a formula-based model to a matrix interface
#'
#' Wraps a model function that expects \code{formula + data} so it can be called
#' with a plain \code{X} (data.frame or matrix) and \code{y} (vector).
#' Factors in \code{X} are preserved; special column names are safely backtick-quoted
#' in the generated formula so they survive the formula parser.
#'
#' @param fit_func   A model-fitting function whose first two arguments are
#'   \code{formula} and \code{data} (e.g. \code{lm}, \code{glm}).
#' @param predict_func A prediction function with signature
#'   \code{function(model, newdata, ...)}. Defaults to \code{stats::predict}.
#'
#' @return A named list with two elements:
#' \describe{
#'   \item{\code{fit(X, y, weights, ...)}}{Fits the model. \code{X} is a
#'     data.frame (or coercible matrix), \code{y} is the response vector.
#'     Extra arguments are forwarded to \code{fit_func}.}
#'   \item{\code{predict(model, newdata, ...)}}{Generates predictions.
#'     \code{newdata} must have the same columns as the \code{X} used in
#'     \code{fit}. Extra arguments are forwarded to \code{predict_func}.}
#' }
#'
#' @examples
#' lm_matrix <- formula_to_matrix(lm)
#' X <- data.frame(wt = mtcars$wt, hp = mtcars$hp, cyl = factor(mtcars$cyl))
#' y <- mtcars$mpg
#' model <- lm_matrix$fit(X, y)
#' lm_matrix$predict(model, X[1:5, ])
#'
#' @export
formula_to_matrix <- function(fit_func, predict_func = stats::predict) {
  
  ## ---- helpers ---------------------------------------------------------------
  
  # Safely quote a single column name for use inside a formula string.
  # Names that are valid R symbols are left as-is; everything else is
  # wrapped in backticks (with any embedded backticks escaped).
  safe_quote <- function(nm) {
    if (make.names(nm) == nm) nm
    else paste0("`", gsub("`", "\\\\`", nm), "`")
  }
  
  # Build a formula of the form:  .response ~ col1 + col2 + ...
  # using a dedicated response column injected into the data.
  build_formula <- function(col_names) {
    rhs <- if (length(col_names) == 0L) "1"
    else paste(vapply(col_names, safe_quote, character(1L)), collapse = " + ")
    stats::as.formula(paste(".response ~", rhs))
  }
  
  ## ---- fit -------------------------------------------------------------------
  
  fit <- function(X, y, weights = NULL, ...) {
    if (!is.data.frame(X)) X <- as.data.frame(X)
    
    # Inject response into a *copy* of X so the original is unmodified.
    df        <- X
    df[[".response"]] <- y
    
    fml <- build_formula(names(X))
    
    # lm() (and many formula-based models) capture the `weights` argument as
    # a language object via match.call(). Passing it through `...` or as a
    # named argument from inside a wrapper causes "invalid type (closure)"
    # errors.  The safe pattern is to inject weights as a column in `df` and
    # reference it by name.
    extra <- list(...)
    if (!is.null(weights)) {
      df[[".weights"]] <- weights
      do.call(fit_func, c(list(fml, data = df, weights = df[[".weights"]]), extra))
    } else {
      do.call(fit_func, c(list(fml, data = df), extra))
    }
  }
  
  ## ---- predict ---------------------------------------------------------------
  
  predict <- function(model, newdata, ...) {
    if (!is.data.frame(newdata)) newdata <- as.data.frame(newdata)
    predict_func(model, newdata = newdata, ...)
  }
  
  list(fit = fit, predict = predict)
}


#' Convert a matrix-based model to a formula interface
#'
#' Wraps a model function that expects a numeric matrix \code{X} and a response
#' vector \code{y} (like \code{glmnet::glmnet}) so it can be called with the
#' familiar \code{formula + data} interface.  The formula is expanded via
#' \code{\link[stats]{model.matrix}}, which handles factor dummy-coding,
#' interactions, and inline transformations automatically.
#'
#' @param fit_func     A model-fitting function whose first two positional
#'   arguments are \code{x} (numeric matrix) and \code{y} (response vector),
#'   e.g. \code{glmnet::glmnet}.
#' @param predict_func A prediction function with signature
#'   \code{function(model, newX, ...)} where \code{newX} is a numeric matrix.
#'   Defaults to a thin wrapper around \code{stats::predict} that passes
#'   \code{newdata} as \code{newx}.
#'
#' @return A named list with two elements:
#' \describe{
#'   \item{\code{fit(formula, data, ...)}}{Fits the model. The formula is
#'     expanded with \code{model.matrix}; the intercept column is dropped
#'     before passing to \code{fit_func} (add it back via \code{...} if your
#'     model needs it). Extra arguments are forwarded to \code{fit_func}.}
#'   \item{\code{predict(model, newdata, ...)}}{Generates predictions.
#'     \code{newdata} is expanded with the same \code{model.matrix} terms
#'     captured at fit time. Extra arguments are forwarded to
#'     \code{predict_func}.}
#' }
#'
#' @examples
#' \dontrun{
#' glmnet_formula <- matrix_to_formula(
#'   fit_func = glmnet::glmnet,
#'   predict_func = function(model, newX, ...) {
#'     glmnet::predict.glmnet(model, newx = newX, s = 0.01, ...)
#'   }
#' )
#' model <- glmnet_formula$fit(mpg ~ wt + hp + factor(cyl), data = mtcars)
#' glmnet_formula$predict(model, newdata = mtcars[1:5, ])
#' }
#'
#' @export
matrix_to_formula <- function(
    fit_func,
    predict_func = function(model, newX, ...) stats::predict(model, newdata = newX, ...)
) {
  
  # We store the formula terms object after fitting so that newdata is expanded
  # identically (same dummy levels, same column order) at predict time.
  terms_ref <- NULL   # will be set inside fit(), read inside predict()
  
  ## ---- fit -------------------------------------------------------------------
  
  fit <- function(formula, data, ...) {
    mf   <- stats::model.frame(formula, data = data)
    mt   <- attr(mf, "terms")
    y    <- stats::model.response(mf)
    
    # model.matrix drops the response but keeps the intercept column "(Intercept)".
    # Most matrix-interface models (glmnet, xgboost, …) want no intercept column.
    X    <- stats::model.matrix(mt, data = mf)
    X    <- X[, colnames(X) != "(Intercept)", drop = FALSE]
    
    # Persist terms so predict() can re-expand newdata identically.
    terms_ref <<- mt
    
    fit_func(X, y, ...)
  }
  
  ## ---- predict ---------------------------------------------------------------
  
  predict <- function(model, newdata, ...) {
    if (is.null(terms_ref)) {
      stop("Call `fit()` before `predict()`.")
    }
    
    newX <- stats::model.matrix(terms_ref, data = newdata)
    newX <- newX[, colnames(newX) != "(Intercept)", drop = FALSE]
    
    predict_func(model, newX, ...)
  }
  
  list(fit = fit, predict = predict)
}

test_that("Model initialization works", {
  # Valid initialization
  mod <- Model$new(glmnet::glmnet)
  expect_s3_class(mod, "Model")
  expect_s3_class(mod, "R6")
  expect_null(mod$fitted)
  expect_null(mod$task)
  
  # Invalid initialization
  expect_error(Model$new("not_a_function"))
  expect_error(Model$new(123))
})

---
title: "Use `unifiedml` for benchmarking models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Use `unifiedml` for benchmarking models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Classification

```{r fig.width=7, eval=TRUE}
library(unifiedml)
library(randomForest)
library(e1071)
library(caret)
set.seed(123)

X <- iris[, 1:4]
y <- iris$Species

models <- list(
  glm  = Model$new(caret::train),
  rf   = Model$new(randomForest::randomForest),
  svm  = Model$new(e1071::svm)   
)

params <- list(
  glm = list(method = "glmnet",
             tuneGrid = data.frame(alpha = 0, lambda = 0.01),
             trControl = trainControl(method = "none")),
  
  rf  = list(ntree = 150),
  
  svm = list(kernel = "radial",   # <-- added
             cost = 1,
             gamma = 0.1)
)

results <- benchmark(models, X, y, cv = 5, params = params)
print(results)
```

# Regression 

```{r, eval=TRUE}
library(unifiedml)
library(randomForest)
library(e1071)
library(caret)

set.seed(123)

# Regression data
X <- mtcars[, setdiff(names(mtcars), "mpg")]
y <- mtcars$mpg

models <- list(
  glm  = Model$new(caret::train),
  rf   = Model$new(randomForest::randomForest),
  svm  = Model$new(e1071::svm)
)

params <- list(
  glm = list(method = "glmnet",
             tuneGrid = data.frame(alpha = 0, lambda = 0.01),
             trControl = trainControl(method = "none")),
  
  rf  = list(ntree = 150),
  
  svm = list(type = "eps-regression",  # <-- important for regression
             kernel = "radial",
             cost = 1,
             gamma = 0.1)
)

results <- benchmark(models, X, y, cv = 5, params = params)
print(results)
```

---
title: "Predict proba"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Predict proba}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# 2 - `iris` data

```{r fig.width=7}
# ============================================================================
# WORKING EXAMPLES: predict_proba with unifiedml using IRIS dataset
# ============================================================================

# Load required packages
library(unifiedml)
library(randomForest)
library(nnet)
library(e1071)

# Load iris dataset
data(iris)

# Setup reproducible data
set.seed(42)

# Create feature matrix (all 4 numeric features)
X <- as.matrix(iris[, 1:4])
colnames(X) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")

# Target: Species (multi-class with 3 levels)
y_multiclass <- iris$Species

# Create binary classification target (Versicolor vs others)
y_binary <- factor(
  ifelse(iris$Species == "versicolor", "versicolor", "other"),
  levels = c("other", "versicolor")
)

# Split into train/test (75% train, 25% test)
set.seed(42)
train_idx <- sample(1:nrow(X), size = floor(0.75 * nrow(X)), replace = FALSE)
test_idx <- setdiff(1:nrow(X), train_idx)

X_train <- X[train_idx, ]
X_test <- X[test_idx, ]
y_train_multiclass <- y_multiclass[train_idx]
y_test_multiclass <- y_multiclass[test_idx]
y_train_binary <- y_binary[train_idx]
y_test_binary <- y_binary[test_idx]

cat("\n")
cat("============================================================================\n")
cat("IRIS DATASET - Summary\n")
cat("============================================================================\n")
cat(sprintf("Training samples: %d\n", nrow(X_train)))
cat(sprintf("Test samples: %d\n", nrow(X_test)))
cat(sprintf("Features: %d\n", ncol(X_train)))
cat(sprintf("Classes: %s\n", paste(levels(y_multiclass), collapse = ", ")))

# ============================================================================
# EXAMPLE 1: randomForest - Multi-class Classification on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 1: randomForest - Multi-class Classification\n")
cat("============================================================================\n")

mod_rf <- Model$new(randomForest::randomForest)
mod_rf$fit(X_train, y_train_multiclass, ntree = 100)

cat("\nPredicting probabilities for first 5 test samples:\n")
probs_rf <- mod_rf$predict_proba(X_test[1:5, ])

cat("\nProbability matrix:\n")
print(round(probs_rf, 3))

cat("\nInterpretation:\n")
for(i in 1:5) {
  cat(sprintf("\nSample %d (Actual: %s):\n", i, as.character(y_test_multiclass[i])))
  cat(sprintf("  setosa:     %.1f%%\n", probs_rf[i, "setosa"] * 100))
  cat(sprintf("  versicolor: %.1f%%\n", probs_rf[i, "versicolor"] * 100))
  cat(sprintf("  virginica:  %.1f%%\n", probs_rf[i, "virginica"] * 100))
  cat(sprintf("  Predicted:  %s\n", colnames(probs_rf)[which.max(probs_rf[i, ])]))
}

# Get class predictions
pred_classes_rf <- mod_rf$predict(X_test[1:5, ], type = "class")
cat("\nPredicted classes (first 5):", as.character(pred_classes_rf), "\n")
cat("Actual classes (first 5):   ", as.character(y_test_multiclass[1:5]), "\n")

# Calculate accuracy on full test set
probs_all_rf <- mod_rf$predict_proba(X_test)
pred_all_rf <- colnames(probs_all_rf)[apply(probs_all_rf, 1, which.max)]
accuracy_rf <- mean(pred_all_rf == as.character(y_test_multiclass))
cat(sprintf("\nTest set accuracy: %.1f%%\n", accuracy_rf * 100))

# ============================================================================
# EXAMPLE 2: nnet - Multi-class Classification on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 2: nnet - Multi-class Classification\n")
cat("============================================================================\n")

mod_nnet <- Model$new(nnet::nnet)
mod_nnet$fit(X_train, y_train_multiclass, size = 10, maxit = 200, trace = FALSE)

cat("\nPredicting probabilities for first 5 test samples:\n")
probs_nnet <- mod_nnet$predict_proba(X_test[1:5, ])

cat("\nProbability matrix (all 3 classes):\n")
print(round(probs_nnet, 3))

cat("\nDetailed predictions:\n")
for(i in 1:5) {
  cat(sprintf("\nSample %d (Actual: %s):\n", i, as.character(y_test_multiclass[i])))
  cat(sprintf("  setosa:     %.1f%%\n", probs_nnet[i, "setosa"] * 100))
  cat(sprintf("  versicolor: %.1f%%\n", probs_nnet[i, "versicolor"] * 100))
  cat(sprintf("  virginica:  %.1f%%\n", probs_nnet[i, "virginica"] * 100))
  cat(sprintf("  Predicted:  %s\n", colnames(probs_nnet)[which.max(probs_nnet[i, ])]))
}

# Get class predictions
pred_classes_nnet <- mod_nnet$predict(X_test[1:5, ], type = "class")
cat("\nPredicted classes (first 5):", as.character(pred_classes_nnet), "\n")
cat("Actual classes (first 5):   ", as.character(y_test_multiclass[1:5]), "\n")

# Calculate accuracy
probs_all_nnet <- mod_nnet$predict_proba(X_test)
pred_all_nnet <- colnames(probs_all_nnet)[apply(probs_all_nnet, 1, which.max)]
accuracy_nnet <- mean(pred_all_nnet == as.character(y_test_multiclass))
cat(sprintf("\nTest set accuracy: %.1f%%\n", accuracy_nnet * 100))

# ============================================================================
# EXAMPLE 3: SVM - Multi-class Classification on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 3: SVM - Multi-class Classification\n")
cat("============================================================================\n")

mod_svm <- Model$new(e1071::svm)
mod_svm$fit(X_train, y_train_multiclass, probability = TRUE, kernel = "radial")

cat("\nPredicting probabilities for first 5 test samples:\n")
probs_svm <- mod_svm$predict_proba(X_test[1:5, ])

cat("\nProbability matrix:\n")
print(round(probs_svm, 4))

cat("\nDetailed predictions:\n")
for(i in 1:5) {
  cat(sprintf("\nSample %d (Actual: %s):\n", i, as.character(y_test_multiclass[i])))
  cat(sprintf("  setosa:     %.1f%%\n", probs_svm[i, "setosa"] * 100))
  cat(sprintf("  versicolor: %.1f%%\n", probs_svm[i, "versicolor"] * 100))
  cat(sprintf("  virginica:  %.1f%%\n", probs_svm[i, "virginica"] * 100))
  cat(sprintf("  Predicted:  %s\n", colnames(probs_svm)[which.max(probs_svm[i, ])]))
}

# Calculate accuracy
probs_all_svm <- mod_svm$predict_proba(X_test)
pred_all_svm <- colnames(probs_all_svm)[apply(probs_all_svm, 1, which.max)]
accuracy_svm <- mean(pred_all_svm == as.character(y_test_multiclass))
cat(sprintf("\nTest set accuracy: %.1f%%\n", accuracy_svm * 100))

# ============================================================================
# EXAMPLE 4: Binary Classification on IRIS (Versicolor vs others)
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 4: Binary Classification - Versicolor vs Others\n")
cat("============================================================================\n")

# randomForest binary
mod_rf_binary <- Model$new(randomForest::randomForest)
mod_rf_binary$fit(X_train, y_train_binary, ntree = 100)

cat("\nrandomForest - Binary probabilities (first 5 test samples):\n")
probs_rf_binary <- mod_rf_binary$predict_proba(X_test[1:5, ])
print(round(probs_rf_binary, 3))

# SVM binary
mod_svm_binary <- Model$new(e1071::svm)
mod_svm_binary$fit(X_train, y_train_binary, probability = TRUE, kernel = "radial")

cat("\nSVM - Binary probabilities (first 5 test samples):\n")
probs_svm_binary <- mod_svm_binary$predict_proba(X_test[1:5, ])
print(round(probs_svm_binary, 4))

# Compare binary predictions
cat("\nComparison of Versicolor probabilities:\n")
comparison_binary <- data.frame(
  Sample = 1:5,
  Actual = as.character(y_test_binary[1:5]),
  RandomForest = round(probs_rf_binary[, "versicolor"], 3),
  SVM = round(probs_svm_binary[, "versicolor"], 4)
)
print(comparison_binary)

# ============================================================================
# EXAMPLE 5: Using unified predict() method on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 5: Using unified predict() method\n")
cat("============================================================================\n")

cat("\nrandomForest - predict(type='prob') on first 3 samples:\n")
print(round(mod_rf$predict(X_test[1:3, ], type = "prob"), 3))

cat("\nrandomForest - predict(type='class') on first 3 samples:\n")
print(mod_rf$predict(X_test[1:3, ], type = "class"))

cat("\nnnet - predict(type='class') on first 3 samples:\n")
print(mod_nnet$predict(X_test[1:3, ], type = "class"))

cat("\nSVM - predict(type='class') on first 3 samples:\n")
print(mod_svm$predict(X_test[1:3, ], type = "class"))

# ============================================================================
# EXAMPLE 6: Model Comparison on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 6: Model Performance Comparison\n")
cat("============================================================================\n")

# Compare accuracies
cat("\nModel Accuracies on IRIS test set:\n")
cat(sprintf("  randomForest: %.1f%%\n", accuracy_rf * 100))
cat(sprintf("  nnet:         %.1f%%\n", accuracy_nnet * 100))
cat(sprintf("  SVM:          %.1f%%\n", accuracy_svm * 100))

# Compare predictions for specific samples
cat("\nDetailed comparison for first 5 test samples:\n")
comparison_multi <- data.frame(
  Sample = 1:5,
  Actual = as.character(y_test_multiclass[1:5]),
  RF_Pred = as.character(mod_rf$predict(X_test[1:5, ], type = "class")),
  nnet_Pred = as.character(mod_nnet$predict(X_test[1:5, ], type = "class")),
  SVM_Pred = as.character(mod_svm$predict(X_test[1:5, ], type = "class"))
)
print(comparison_multi)

# ============================================================================
# EXAMPLE 7: Confidence Analysis on IRIS
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 7: Prediction Confidence Analysis\n")
cat("============================================================================\n")

# randomForest confidence
rf_confidences <- apply(probs_all_rf, 1, max)
cat("\nrandomForest - Prediction confidence:\n")
cat(sprintf("  Mean confidence: %.1f%%\n", mean(rf_confidences) * 100))
cat(sprintf("  Median confidence: %.1f%%\n", median(rf_confidences) * 100))
cat(sprintf("  Low confidence (<70%%): %d samples (%.1f%%)\n", 
            sum(rf_confidences < 0.7), mean(rf_confidences < 0.7) * 100))
cat(sprintf("  High confidence (>90%%): %d samples (%.1f%%)\n", 
            sum(rf_confidences > 0.9), mean(rf_confidences > 0.9) * 100))

# nnet confidence
nnet_confidences <- apply(probs_all_nnet, 1, max)
cat("\nnnet - Prediction confidence:\n")
cat(sprintf("  Mean confidence: %.1f%%\n", mean(nnet_confidences) * 100))
cat(sprintf("  Median confidence: %.1f%%\n", median(nnet_confidences) * 100))
cat(sprintf("  Low confidence (<70%%): %d samples (%.1f%%)\n", 
            sum(nnet_confidences < 0.7), mean(nnet_confidences < 0.7) * 100))
cat(sprintf("  High confidence (>90%%): %d samples (%.1f%%)\n", 
            sum(nnet_confidences > 0.9), mean(nnet_confidences > 0.9) * 100))

# SVM confidence
svm_confidences <- apply(probs_all_svm, 1, max)
cat("\nSVM - Prediction confidence:\n")
cat(sprintf("  Mean confidence: %.1f%%\n", mean(svm_confidences) * 100))
cat(sprintf("  Median confidence: %.1f%%\n", median(svm_confidences) * 100))
cat(sprintf("  Low confidence (<70%%): %d samples (%.1f%%)\n", 
            sum(svm_confidences < 0.7), mean(svm_confidences < 0.7) * 100))
cat(sprintf("  High confidence (>90%%): %d samples (%.1f%%)\n", 
            sum(svm_confidences > 0.9), mean(svm_confidences > 0.9) * 100))

# ============================================================================
# EXAMPLE 8: Misclassification Analysis
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("EXAMPLE 8: Misclassification Analysis (randomForest)\n")
cat("============================================================================\n")

# Find misclassified samples
rf_misclassified <- which(pred_all_rf != as.character(y_test_multiclass))

if(length(rf_misclassified) > 0) {
  cat(sprintf("\nFound %d misclassified samples:\n", length(rf_misclassified)))
  
  for(idx in rf_misclassified[1:min(3, length(rf_misclassified))]) {
    cat(sprintf("\nSample %d:\n", idx))
    cat(sprintf("  True class: %s\n", as.character(y_test_multiclass[idx])))
    cat(sprintf("  Predicted: %s\n", pred_all_rf[idx]))
    cat("  Probabilities:\n")
    cat(sprintf("    setosa:     %.1f%%\n", probs_all_rf[idx, "setosa"] * 100))
    cat(sprintf("    versicolor: %.1f%%\n", probs_all_rf[idx, "versicolor"] * 100))
    cat(sprintf("    virginica:  %.1f%%\n", probs_all_rf[idx, "virginica"] * 100))
  }
} else {
  cat("\nPerfect classification! No misclassified samples.\n")
}

# ============================================================================
# SUMMARY
# ============================================================================

cat("\n")
cat("============================================================================\n")
cat("SUMMARY - IRIS Dataset\n")
cat("============================================================================\n")

cat("
✓ SUCCESSFUL EXAMPLES WITH IRIS DATASET:
  1. randomForest - Multi-class classification (3 species)
  2. nnet - Multi-class classification
  3. SVM - Multi-class classification with probabilities
  4. Binary classification (Versicolor vs others)
  5. Unified predict() interface
  6. Model comparison and accuracy analysis
  7. Confidence analysis
  8. Misclassification analysis

✓ KEY FINDINGS ON IRIS:
  • All models achieve high accuracy (>90%) on iris dataset
  • SVM tends to produce extreme probabilities (near 0 or 1)
  • randomForest and nnet show more calibrated probabilities
  • Setosa is perfectly separable from other species
  • Confusion typically occurs between versicolor and virginica

✓ predict_proba() FEATURES DEMONSTRATED:
  • Returns matrix [n_samples × 3] for multi-class
  • Column names: setosa, versicolor, virginica
  • All rows sum to 1
  • Works seamlessly across all model types

All working examples on IRIS dataset completed successfully!\n")
```

---
title: "Matrix or Formula interface"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Matrix or Formula interface}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
library(unifiedml)
```

```{r}
# Example 1: lm with matrix interface (including factors)
demo_lm_matrix <- function() {
  cat("\n=== Example 1: lm with matrix interface ===\n")
  
  lm_matrix <- unifiedml::formula_to_matrix(stats::lm)  # Uses generic predict()
  
  # Create data with factor and numeric columns
  X <- data.frame(
    wt = mtcars$wt,
    hp = mtcars$hp,
    cyl = factor(mtcars$cyl)
  )
  y <- mtcars$mpg
  weights <- rep(1, nrow(X))
  
  model <- lm_matrix$fit(X, y, weights = weights)
  preds <- lm_matrix$predict(model, X[1:5, ])
  
  cat("Predictions:\n")
  print(preds)
  cat("\nCoefficients:\n")
  print(coef(model))
}

# Example 2: glmnet with formula interface (with factors)
demo_glmnet_formula <- function() {
  cat("\n=== Example 2: glmnet with formula interface ===\n")
  
  if (!requireNamespace("glmnet", quietly = TRUE)) {
    cat("glmnet package not installed, skipping example\n")
    return(invisible(NULL))
  }
  
  glmnet_formula <- unifiedml::matrix_to_formula(
    fit_func = glmnet::glmnet,
    predict_func = function(model, newX, ...) {
      # Wrapper to provide default s parameter
      glmnet::predict.glmnet(model, newx = newX, s = 0.01, ...)
    }
  )
  
  # Formula with factor - model.matrix will auto-create dummies
  model <- glmnet_formula$fit(mpg ~ wt + hp + factor(cyl), data = mtcars)
  preds <- glmnet_formula$predict(model, newdata = mtcars[1:5, ])
  
  cat("Predictions:\n")
  print(preds)
}

# Example 3: Special characters in column names
demo_special_names <- function() {
  cat("\n=== Example 3: Column names with special characters ===\n")
  
  lm_matrix <- unifiedml::formula_to_matrix(lm)
  
  # Create problematic column names
  X <- data.frame(
    `x-1` = mtcars$wt,      # Dash
    `a:b` = mtcars$hp,       # Colon
    `log(x)` = log(mtcars$disp),  # Parentheses
    check.names = FALSE
  )
  y <- mtcars$mpg
  
  model <- lm_matrix$fit(X, y)
  preds <- lm_matrix$predict(model, X[1:3, ])
  
  cat("Predictions:\n")
  print(preds)
  cat("\nModel handled special column names correctly!\n")
}

# Run examples (uncomment to test)
demo_lm_matrix()
demo_glmnet_formula()
demo_special_names()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/probs_classif.R
\name{extract_probabilities}
\alias{extract_probabilities}
\title{Extract probability predictions from any R model in a standardised format}
\usage{
extract_probabilities(model, X, y_train = NULL, verbose = FALSE)
}
\arguments{
\item{model}{Fitted model object (any class)}

\item{X}{Feature matrix or data.frame for predictions}

\item{y_train}{Optional training labels used to name output columns}

\item{verbose}{Print diagnostic information (default: FALSE)}
}
\value{
Numeric matrix of shape n_samples x n_classes with column names
as class labels. Attributes \code{extraction_method} and
\code{model_class} record how predictions were obtained.
}
\description{
Extract probability predictions from any R model in a standardised format
}

\name{unifiedml-package}
\alias{unifiedml-package}
\alias{unifiedml}
\docType{package}
\title{\packageTitle{unifiedml}}
\description{\packageDescription{unifiedml}}
\section{Package Content}{\packageIndices{unifiedml}}
\author{\packageAuthor{unifiedml}}
\section{Maintainer}{\packageMaintainer{unifiedml}}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cross_val_score.R
\name{cross_val_score}
\alias{cross_val_score}
\title{Cross-Validation for Model Objects}
\usage{
cross_val_score(
  model,
  X,
  y,
  cv = 5,
  scoring = NULL,
  show_progress = TRUE,
  verbose = TRUE,
  cl = NULL,
  seed = 123,
  fit_params = NULL,
  predict_params = NULL
)
}
\arguments{
\item{model}{A Model object}

\item{X}{Feature matrix or data.frame}

\item{y}{Target vector (type determines regression vs classification)}

\item{cv}{Number of cross-validation folds (default: 5)}

\item{scoring}{Scoring metric: "rmse", "mae", "accuracy", "f1", or a
custom function with signature \code{function(true, pred)} returning
a scalar. Default: auto-detected based on task type.}

\item{show_progress}{Whether to show progress bar (default: TRUE) in sequential mode}

\item{verbose}{logical flag enabling verbose messages (default: TRUE) in parallel mode}

\item{cl}{Optional number of clusters for parallel processing
If using \code{cl} for parallel execution, custom scoring functions must
be self-contained (no dependencies on the calling environment).}

\item{seed}{Reproducibility seed}

\item{fit_params}{A list of additional arguments passed to model$fit()}

\item{predict_params}{A list of additional arguments passed to model$predict()}
}
\value{
Vector of cross-validation scores for each fold
}
\description{
Perform k-fold cross-validation with consistent scoring metrics
across different model types. The scoring metric is automatically
selected based on the detected task type.
}
\examples{
\dontrun{
library(glmnet)
X <- matrix(rnorm(100), ncol = 4)
y <- 2*X[,1] - 1.5*X[,2] + rnorm(25)  # numeric -> regression

mod <- Model$new(glmnet::glmnet)
(cv_scores <- cross_val_score(mod, X, y, cv = 5))  # auto-uses RMSE
mean(cv_scores)  # Average RMSE

cross_val_score(mod, X, y,
fit_params     = list(alpha = 0, lambda = 0.1),
predict_params = list(type = "response"))

cross_val_score(mod, X, y,
fit_params     = list(alpha = 0.5, lambda = 0.1),
predict_params = list(type = "response"))

# Custom scoring: R-squared
r2 <- function(true, pred) {
  ss_res <- sum((true - pred)^2)
  ss_tot <- sum((true - mean(true))^2)
  1 - ss_res / ss_tot
}

(cv_scores4 <- cross_val_score(mod, X, y, cv = 5, scoring = r2))
mean(cv_scores4)  # Average R²

# Classification with accuracy scoring
data(iris)
X_class <- iris[, 1:4]
y_class <- iris$Species  # factor -> classification
mod2 <- Model$new(e1071::svm)
(cv_scores2 <- cross_val_score(mod2, X_class, y_class, cv = 5))  # auto-uses accuracy
mean(cv_scores2)  # Average accuracy

iris_bin <- iris[iris$Species != "virginica", ]
X_bin <- iris_bin[, 1:4]
y_bin <- droplevels(iris_bin$Species)
(cv_scores3 <- cross_val_score(mod2, X_bin, y_bin, cv = 3, 
scoring="f1", fit_params=list(kernel="polynomial")))  
mean(cv_scores3)  # Average F1
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interface_to_interface.R
\name{matrix_to_formula}
\alias{matrix_to_formula}
\title{Convert a matrix-based model to a formula interface}
\usage{
matrix_to_formula(
  fit_func,
  predict_func = function(model, newX, ...) stats::predict(model, newdata = newX, ...)
)
}
\arguments{
\item{fit_func}{A model-fitting function whose first two positional
arguments are \code{x} (numeric matrix) and \code{y} (response vector),
e.g. \code{glmnet::glmnet}.}

\item{predict_func}{A prediction function with signature
\code{function(model, newX, ...)} where \code{newX} is a numeric matrix.
Defaults to a thin wrapper around \code{stats::predict} that passes
\code{newdata} as \code{newx}.}
}
\value{
A named list with two elements:
\describe{
\item{\code{fit(formula, data, ...)}}{Fits the model. The formula is
expanded with \code{model.matrix}; the intercept column is dropped
before passing to \code{fit_func} (add it back via \code{...} if your
model needs it). Extra arguments are forwarded to \code{fit_func}.}
\item{\code{predict(model, newdata, ...)}}{Generates predictions.
\code{newdata} is expanded with the same \code{model.matrix} terms
captured at fit time. Extra arguments are forwarded to
\code{predict_func}.}
}
}
\description{
Wraps a model function that expects a numeric matrix \code{X} and a response
vector \code{y} (like \code{glmnet::glmnet}) so it can be called with the
familiar \code{formula + data} interface.  The formula is expanded via
\code{\link[stats]{model.matrix}}, which handles factor dummy-coding,
interactions, and inline transformations automatically.
}
\examples{
\dontrun{
glmnet_formula <- matrix_to_formula(
  fit_func = glmnet::glmnet,
  predict_func = function(model, newX, ...) {
    glmnet::predict.glmnet(model, newx = newX, s = 0.01, ...)
  }
)
model <- glmnet_formula$fit(mpg ~ wt + hp + factor(cyl), data = mtcars)
glmnet_formula$predict(model, newdata = mtcars[1:5, ])
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interface_to_interface.R
\name{formula_to_matrix}
\alias{formula_to_matrix}
\title{Convert a formula-based model to a matrix interface}
\usage{
formula_to_matrix(fit_func, predict_func = stats::predict)
}
\arguments{
\item{fit_func}{A model-fitting function whose first two arguments are
\code{formula} and \code{data} (e.g. \code{lm}, \code{glm}).}

\item{predict_func}{A prediction function with signature
\code{function(model, newdata, ...)}. Defaults to \code{stats::predict}.}
}
\value{
A named list with two elements:
\describe{
\item{\code{fit(X, y, weights, ...)}}{Fits the model. \code{X} is a
data.frame (or coercible matrix), \code{y} is the response vector.
Extra arguments are forwarded to \code{fit_func}.}
\item{\code{predict(model, newdata, ...)}}{Generates predictions.
\code{newdata} must have the same columns as the \code{X} used in
\code{fit}. Extra arguments are forwarded to \code{predict_func}.}
}
}
\description{
Wraps a model function that expects \code{formula + data} so it can be called
with a plain \code{X} (data.frame or matrix) and \code{y} (vector).
Factors in \code{X} are preserved; special column names are safely backtick-quoted
in the generated formula so they survive the formula parser.
}
\examples{
lm_matrix <- formula_to_matrix(lm)
X <- data.frame(wt = mtcars$wt, hp = mtcars$hp, cyl = factor(mtcars$cyl))
y <- mtcars$mpg
model <- lm_matrix$fit(X, y)
lm_matrix$predict(model, X[1:5, ])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/benchmark.R
\name{benchmark}
\alias{benchmark}
\title{Benchmark Multiple Models with Cross-Validation and Model-Specific Parameters}
\usage{
benchmark(
  models,
  X,
  y,
  cv = 5L,
  scoring = NULL,
  params = NULL,
  cl = NULL,
  show_progress = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{models}{A named list of \code{Model$new(...)} objects to benchmark.}

\item{X}{A data frame or matrix of predictors.}

\item{y}{A vector of outcomes (factor for classification, numeric for regression).}

\item{cv}{Integer, number of cross-validation folds (default 5).}

\item{scoring}{Scoring metric: "rmse", "mae", "accuracy", or "f1"
(default: auto-detected based on task)}

\item{params}{Optional named list of lists, each sublist containing extra arguments
to pass to the corresponding model's \code{fit()} call. Names must match \code{models}.}

\item{cl}{Optional number of clusters for parallel processing}

\item{show_progress}{Logical, whether to show a progress bar (default TRUE).}

\item{verbose}{Logical, whether to print messages about each model (default TRUE).}
}
\value{
A list containing the CV scores for each model.
}
\description{
Perform k-fold cross-validation on a list of models, using model-specific parameters.
Supports verbose messages and a progress bar.
}
\examples{
\dontrun{
library(randomForest)

X <- iris[, 1:4]
y <- iris$Species

models <- list(
  glm  = Model$new(caret::train),
  rf   = Model$new(randomForest::randomForest),
  xgb  = Model$new(caret::train)
)

params <- list(
  glm = list(method = "glmnet",
             tuneGrid = data.frame(alpha = 0, lambda = 0.01),
             trControl = trainControl(method = "none")),
  rf  = list(ntree = 150),
  xgb = list(method = "xgbTree",
             tuneGrid = data.frame(nrounds = 150, max_depth = 3, eta = 0.3,
                                   gamma = 0, colsample_bytree = 1,
                                   min_child_weight = 1, subsample = 1),
             trControl = trainControl(method = "none"))
)

results <- benchmark(models, X, y, cv = 5, params = params,
                     show_progress = TRUE, verbose = TRUE)
print(results)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\docType{class}
\name{Model}
\alias{Model}
\title{Unified Machine Learning Interface using R6}
\description{
Provides a consistent interface for various machine learning models in R,
with automatic detection of formula vs matrix interfaces, built-in
cross-validation, model interpretability, and visualization.

An R6 class that provides a unified interface for regression and classification
models with automatic interface detection, cross-validation, and interpretability
features. The task type (regression vs classification) is automatically detected
from the response variable type.
}
\examples{
\donttest{
# Regression example with glmnet
library(glmnet)
X <- matrix(rnorm(100), ncol = 4)
y <- 2*X[,1] - 1.5*X[,2] + rnorm(25)  # numeric -> regression

mod <- Model$new(glmnet::glmnet)
mod$fit(X, y, alpha = 0, lambda = 0.1)
mod$summary()
predictions <- mod$predict(X)

# Classification example  
data(iris)
iris_binary <- iris[iris$Species \%in\% c("setosa", "versicolor"), ]
X_class <- as.matrix(iris_binary[, 1:4])
y_class <- droplevels(iris_binary$Species)  # factor -> classification

mod2 <- Model$new(e1071::svm)
mod2$fit(X_class, y_class, kernel = "radial")
predictions <- mod2$predict(X_class)
mod2$predict_proba(X_class)

}

}
\author{
Your Name
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{model_fn}}{The modeling function (e.g., glmnet::glmnet, randomForest::randomForest)}

\item{\code{fitted}}{The fitted model object}

\item{\code{task}}{Type of task: "regression" or "classification" (automatically detected)}

\item{\code{X_train}}{Training features matrix}

\item{\code{y_train}}{Training target vector}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Model-new}{\code{Model$new()}}
\item \href{#method-Model-fit}{\code{Model$fit()}}
\item \href{#method-Model-predict}{\code{Model$predict()}}
\item \href{#method-Model-predict_proba}{\code{Model$predict_proba()}}
\item \href{#method-Model-print}{\code{Model$print()}}
\item \href{#method-Model-summary}{\code{Model$summary()}}
\item \href{#method-Model-plot}{\code{Model$plot()}}
\item \href{#method-Model-clone_model}{\code{Model$clone_model()}}
\item \href{#method-Model-clone}{\code{Model$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Model-new"></a>}}
\if{latex}{\out{\hypertarget{method-Model-new}{}}}
\subsection{Method \code{new()}}{
Initialize a new Model
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Model$new(model_fn)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{model_fn}}{A modeling function (e.g., glmnet, randomForest, svm)}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new Model object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Model-fit"></a>}}
\if{latex}{\out{\hypertarget{method-Model-fit}{}}}
\subsection{Method \code{fit()}}{
Fit the model to training data

Automatically detects task type (regression vs classification) based on
the type of the response variable y. Numeric y -> regression,
factor y -> classification.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Model$fit(X, y, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{X}}{Feature matrix or data.frame}

\item{\code{y}}{Target vector (numeric for regression, factor for classification)}

\item{\code{...}}{Additional arguments passed to the model function}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
self (invisible) for method chaining
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Model-predict"></a>}}
\if{latex}{\out{\hypertarget{method-Model-predict}{}}}
\subsection{Method \code{predict()}}{
Generate predictions from fitted model
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Model$predict(X, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{X}}{Feature matrix for prediction}

\item{\code{...}}{Additional arguments passed to predict function}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Vector of predictions
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Model-predict_proba"></a>}}
\if{latex}{\out{\hypertarget{method-Model-predict_proba}{}}}
\subsection{Method \code{predict_proba()}}{
Predict probabilities from fitted model
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Model$predict_proba(X)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{X}}{Feature matrix for prediction}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Model-print"></a>}}
\if{latex}{\out{\hypertarget{method-Model-print}{}}}
\subsection{Method \code{print()}}{
Print model information
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Model$print()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
self (invisible) for method chaining
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Model-summary"></a>}}
\if{latex}{\out{\hypertarget{method-Model-summary}{}}}
\subsection{Method \code{summary()}}{
Compute numerical derivatives and statistical significance

Uses finite differences to compute approximate partial derivatives
for each feature, providing model-agnostic interpretability.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Model$summary(h = 0.01, alpha = 0.05)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{h}}{Step size for finite differences (default: 0.01)}

\item{\code{alpha}}{Significance level for p-values (default: 0.05)}
}
\if{html}{\out{</div>}}
}
\subsection{Details}{
The method computes numerical derivatives using central differences.

Statistical significance is assessed using t-tests on the derivative
estimates across samples.
}

\subsection{Returns}{
A data.frame with derivative statistics (invisible)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Model-plot"></a>}}
\if{latex}{\out{\hypertarget{method-Model-plot}{}}}
\subsection{Method \code{plot()}}{
Create partial dependence plot for a feature

Visualizes the relationship between a feature and the predicted outcome
while holding other features at their mean values.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Model$plot(feature = 1, n_points = 100)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{feature}}{Index or name of feature to plot}

\item{\code{n_points}}{Number of points for the grid (default: 100)}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
self (invisible) for method chaining
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Model-clone_model"></a>}}
\if{latex}{\out{\hypertarget{method-Model-clone_model}{}}}
\subsection{Method \code{clone_model()}}{
Create a deep copy of the model

Useful for cross-validation and parallel processing where
multiple independent model instances are needed.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Model$clone_model()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A new Model object with same configuration
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Model-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Model-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Model$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/train_test_split.R
\name{train_test_split}
\alias{train_test_split}
\title{Split data into training and test sets}
\usage{
train_test_split(X, y, test_size = 0.2, seed = NULL)
}
\arguments{
\item{X}{A matrix or data.frame of features.}

\item{y}{A vector of responses (numeric or factor). Must have the same
number of rows as \code{X}.}

\item{test_size}{Proportion of observations to use as the test set.
A number in (0, 1). Default is \code{0.2} (80/20 split).}

\item{seed}{An optional integer random seed for reproducibility. If
\code{NULL} (default) the current RNG state is used.}
}
\value{
A named list with four elements:
\item{X_train}{Training features (same type as \code{X}).}
\item{X_test}{Test features (same type as \code{X}).}
\item{y_train}{Training response.}
\item{y_test}{Test response.}
}
\description{
Randomly splits a feature matrix or data.frame and its corresponding
response vector into training and test subsets.
}
\examples{
# matrix input
X <- iris[, 1:4]
y <- iris$Species
d <- unifiedml::train_test_split(X, y, test_size = 0.3, seed = 42)
dim(d$X_train)  # 105 x 4
dim(d$X_test)   #  45 x 4

# data.frame input
d2 <- unifiedml::train_test_split(iris[, 1:4], iris$Species, test_size = 0.2)
is.data.frame(d2$X_train)  # TRUE

}
